"""
Provides the operator representations needed when building a Hamiltonian representation.
I chose to only implement direct product operators. Not sure if we'll need a 1D base class...
"""

import numpy as np, scipy.sparse as sp, os, tempfile as tf, time, gc, itertools
from collections import OrderedDict
from McUtils.Numputils import SparseArray
import McUtils.Numputils as nput
from McUtils.Scaffolding import Logger, NullLogger
from McUtils.Parallelizers import Parallelizer, SerialNonParallelizer

from .StateSpaces import BasisStateSpace, SelectionRuleStateSpace, PermutationallyReducedStateSpace, BraKetSpace
from .StateFilters import BasisStateSpaceFilter

__all__ = [
    "Operator",
    "ContractedOperator"
]

__reload_hook__ = [ '.StateSpaces' ]

class Operator:
    """
    Provides a (usually) _lazy_ representation of an operator, which allows things like
    QQQ and pQp to be calculated block-by-block.
    Crucially, the underlying basis for the operator is assumed to be orthonormal.
    """
    def __init__(self, funcs, quanta,
                 prod_dim=None,
                 symmetries=None,
                 selection_rules=None,
                 selection_rule_steps=None, # for product operators
                 parallelizer=None,
                 logger=None,
                 zero_threshold=None,
                 skipped_indices=None,
                 chunk_size=None
                 ):
        """
        :param funcs: The functions use to calculate representation
        :type funcs: callable | Iterable[callable]
        :param quanta: The number of quanta to do the deepest-level calculations up to (also tells us dimension)
        :type quanta: int | Iterable[int]
        :param prod_dim: The number of functions in `funcs`, if `funcs` is a direct term generator
        :type prod_dim: int | None
        :param symmetries: Labels for the funcs where if two funcs share a label they are symmetry equivalent
        :type symmetries: Iterable[int] | None
        """
        if isinstance(quanta, int):
            quanta = [quanta]
            # funcs = [funcs]
        try:
            funcs = tuple(funcs)
            single_func = False
        except TypeError:  # can't iterate
            single_func = True
        if prod_dim is None:
            if single_func:
                funcs = (funcs,)
            prod_dim = len(funcs)
        self.fdim = prod_dim
        self.funcs = funcs
        self.selection_rules = selection_rules
        self.selection_rule_steps = selection_rule_steps
        self.symmetry_inds = symmetries
        self.quanta = tuple(quanta)
        self.mode_n = len(quanta)
        self.zero_threshold = 1.0e-14 if zero_threshold is None else zero_threshold
        # self._tensor = None
        if logger is None:
            logger = NullLogger()
        self.logger = logger
        self._parallelizer = parallelizer
        self.chunk_size = chunk_size
        if skipped_indices is not None:
            skipped_indices = {tuple(i) for i in skipped_indices}
        self.skipped_indices = skipped_indices

    def clear_cache(self):
        """
        :return:
        :rtype:
        """

        funcs = self.funcs
        if not isinstance(funcs, tuple):
            funcs = (funcs,)
        for f in funcs:
            if hasattr(f, 'clear_cache'):
                f.clear_cache()

    @property
    def ndim(self):
        return self.fdim + self.mode_n
    @property
    def shape(self):
        return (self.mode_n, ) * self.fdim + self.quanta

    def load_parallelizer(self):
        """
        Loads the held parallelizer if needed

        :return:
        :rtype:
        """
        from McUtils.Parallelizers import Parallelizer

        if not isinstance(self._parallelizer, Parallelizer):
            return Parallelizer.lookup(self._parallelizer)
        else:
            return self._parallelizer

    @property
    def parallelizer(self):
        """
        Loads a parallelizer that can be used to speed up various bits of the calculation

        :return:
        :rtype:
        """
        if self._parallelizer is not None:
            return self.load_parallelizer()
        else:
            return self._parallelizer
    @property
    def is_diagonal(self):
        return self.fdim == 0
    @property
    def is_zero(self):
        return False
    def get_inner_indices(self, reduced_inds=False):
        """
        Gets the n-dimensional array of ijkl (e.g.) indices that functions will map over
        Basically returns the indices of the inner-most tensor

        :return:
        :rtype:
        """
        dims = self.fdim
        if dims == 0:
            return None
        n_terms = max(len(x) for x in self.selection_rules) if reduced_inds else self.mode_n
        shp = (n_terms,) * dims
        inds = np.indices(shp, dtype=int)
        tp = np.roll(np.arange(dims + 1), -1)
        base_tensor = np.transpose(inds, tp)
        return base_tensor

    def __getitem__(self, item):
        # check to see if we were actually given a bunch of bra and ket states
        if isinstance(item, BraKetSpace): # common case, so we use it to skip further fuckery
            pass
        elif (
                isinstance(item, tuple)
                and len(item) == 2
                and all(len(x) == self.mode_n for x in item[0])
        ):
            # expected to look like num_modes X [bra, ket] X quanta
            # so we reshape it
            bras, kets = item
            modes = [None] * self.mode_n
            for i in range(self.mode_n):
                i_bras = [x[i] for x in bras]
                i_kets = [x[i] for x in kets]
                modes[i] = (i_bras, i_kets)
            item = modes
        return self.get_elements(item)

    def __getstate__(self):
        state = self.__dict__.copy()
        state['_parallelizer'] = None
        state['logger'] = None
        state['_tensor'] = None
        return state

    def _get_eye_tensor(self, states):
        """

        :param states:
        :type states: BraKetSpace
        :param quants:
        :type quants:
        :return:
        :rtype:
        """

        ntotal = len(states)
        states, non_orthog = states.apply_non_orthogonality([])
        return sp.csr_matrix(
            (
                np.ones(len(non_orthog)),
                (
                    np.zeros(len(non_orthog)),
                    non_orthog
                )
            ),
            shape=(1, ntotal)
        )


    def filter_symmetric_indices(self, inds):
        """
        Determines which inds are symmetry unique.
        For something like `qqq` all permutations are equivalent, but for `pqp` we have `pi qj pj` distinct from `pj qj pi`.
        This means for `qqq` we have `(1, 0, 0) == (0, 1, 0)` but for `pqp` we only have stuff like `(2, 0, 1) == (1, 0, 2)` .

        :param inds: indices to filter symmetric bits out of
        :type inds: np.ndarray
        :return: symmetric indices & inverse map
        :rtype:
        """

        symm_labels = self.symmetry_inds
        flat = np.reshape(inds, (-1, inds.shape[-1]))
        if symm_labels is None:
            return flat, None

        ind_grps = OrderedDict()
        for i, l in enumerate(symm_labels):
            if l in ind_grps:
                ind_grps[l].append(i)
            else:
                ind_grps[l] = [i]
        ind_grps = list(ind_grps.values())

        if len(ind_grps) == len(symm_labels):
            # short-circuit case if no symmetry
            return flat, None
        elif len(ind_grps) == 1:
            # totally symmetric so we can just sort and delete the dupes
            flat = np.sort(flat, axis=1)
        else:
            # if indices are shared between groups we can't do anything about them
            # so we figure out where there are overlaps in indices
            # we do this iteratively by looping over groups, figuring out pairwise if they
            # share elements, and using an `or` op to update our list of indep terms
            indep_grps = np.full((len(flat),), False, dtype=bool)
            # this is 4D, but over relatively small numbers of inds (maybe 6x6x6x6 at max)
            # and so hopefully still fast, esp. relative to computing actual terms
            for i in range(len(ind_grps)):
                for j in range(i+1, len(ind_grps)):
                    for t1 in ind_grps[i]:
                        for t2 in ind_grps[j]:
                            indep_grps = np.logical_or(
                                indep_grps,
                                flat[:, t1] == flat[:, t2]
                            )
            indep_grps = np.logical_not(indep_grps)
            symmable = flat[indep_grps]

            # pull out the groups of indices and sort them
            symmetriz_grps = [np.sort(symmable[:, g], axis=1) for g in ind_grps]
            # stack them together to build a full array
            # then reshuffle to get the OG ordering
            reordering = np.argsort(np.concatenate(ind_grps))
            symmetrized = np.concatenate(symmetriz_grps, axis=1)[:, reordering]
            flat[indep_grps] = symmetrized

        # next pull the unique indices
        uinds_basic, inverse = np.unique(flat, axis=0, return_inverse=True)

        # Finally apply a lexical sorting so that we
        # can do as little work as possible later since work can be shared between (0,0,0) and (0,0,1) but not
        # (1, 1, 0)

        indsort = np.lexsort(uinds_basic.T)
        # raise Exception(indsort.shape, uinds_basic.shape)
        uinds = uinds_basic[indsort]
        # if len(uinds) > 500:
        #     raise Exception(inds, uinds)

        return uinds, (inverse, np.argsort(indsort))

    def _mat_prod_operator_terms(self, inds, funcs, states, sel_rules):
        """
        Evaluates product operator terms based on 1D representation matrices coming from funcs

        :param inds:
        :type inds:
        :param funcs: functions that generate 1D representation matrices
        :type funcs:
        :param states:
        :type states: BraKetSpace
        :param sel_rules: selection rules as 1D arrays of rules for each dimension
        :type sel_rules:
        :return:
        :rtype:
        """
        from collections import OrderedDict

        # We figure out which indices are actually unique; this gives us a way to map
        # indices onto operators
        # We assume we always have as many indices as dimensions in our operator
        # and so the more unique indices the _lower_ the dimension of the operator
        # since it acts on the same index more times
        # The indices will be mapped onto ints for storage in `pieces`

        uinds = OrderedDict((k, None) for k in inds)
        uinds = np.array(list(uinds.keys()))
        states = states.take_subdimensions(uinds)

        if sel_rules is not None:
            sel_rules = states.get_sel_rules_from1d(inds, sel_rules)
            states, all_sels = states.apply_sel_rules(sel_rules)
        else:
            all_sels = None

        if len(states) == 0:
            return None, None

        bras, kets = states.state_pairs
        max_dim = max(np.max(bras), np.max(kets))
        padding = 3  # for approximation reasons we need to pad our representations...

        mm = {k: i for i, k in enumerate(uinds)}
        subdim = len(uinds)
        # set up a place to hold onto the pieces we'll calculate
        pieces = [None] * subdim
        # now we construct the reps from 1D ones
        for f, i in zip(funcs, inds):
            n = mm[i]  # makes sure that we fill in in the same order as uinds
            if pieces[n] is None:
                pieces[n] = f(max_dim + padding)  # type: sp.spmatrix
            else:
                # QQ -> Q.Q & QQQ -> Q.Q.Q
                og = pieces[n]  # type: sp.spmatrix
                pieces[n] = og.dot(f(max_dim + padding))

        # now we take the requisite products of the chunks for the indices that are
        # potentially non-orthogonal
        chunk = None
        for i, j, o in zip(bras, kets, pieces):
            op = o  # type: sp.spmatrix
            blob = np.asarray(op[i, j]).squeeze()
            if chunk is None:
                if isinstance(blob, (int, np.integer, float, np.floating)):
                    chunk = np.array([blob])
                else:
                    chunk = np.asarray(blob)
            else:
                chunk = chunk * blob

        # weird stuff can happen with sp.spmatrix
        if (
                isinstance(chunk, (int, np.integer, float, np.floating))
                or (isinstance(chunk, np.ndarray) and chunk.ndim == 0)
        ):
            chunk = np.array([chunk])

        return chunk, all_sels

    def _direct_prod_operator_terms(self, inds, func, states):
        """
        Evaluates product operator terms using a function to directly generate them

        :param inds:
        :type inds:
        :param func: a function that will take a set of indices and term-generator
        :type func:
        :return:
        :rtype:
        """

        # next we get the term generator defined by inds
        # this will likely end up calling uinds again, but ah well, that operation is cheap
        uinds = OrderedDict((k, None) for k in inds)
        uinds = np.array(list(uinds.keys()))
        # the issue here is that _this_ operation is not necessarily cheap...
        # the subdimension states still can be quite expensive to get indices from
        states = states.take_subdimensions(uinds)

        gen = func(inds)
        try:
            g, sel_rules = gen
            r = sel_rules[0][0]
            if not isinstance(r, (int, np.integer)):
                raise TypeError("Bad selection rule")
        except (TypeError, IndexError):
            sel_rules = None
        else:
            gen = g
            sel_rules = None

        # bras, kets = states.state_pairs
        # states = [bk for i, bk in enumerate(zip(bras, kets)) if i in inds]
        chunk = gen(zip(*states.state_pairs))

        return chunk, sel_rules

    def _calculate_single_pop_elements(self, inds, funcs, states, sel_rules,
                                       check_orthogonality=True,
                                       use_sel_rule_filtering=False,
                                       use_sel_sum_filtering=False
                                       ):
        """
        Calculates terms for a single product operator.
        Assumes orthogonal product bases.

        :param inds: the index in the total operator tensor (i.e. which of the funcs to use)
        :type inds:
        :param funcs: the functions to use when generating representations, must return matrices
        :type funcs:
        :param states: the states to compute terms between stored like ((s_1l, s_1r), (s_2l, s_2r), ...)
                        where `s_il` is the set of quanta for mode i for the bras and
                        `s_ir` is the set of quanta for mode i for the kets
        :type states: BraKetSpace
        :param quants: the total quanta to use when building representations
        :type quants:
        :param sel_rules: the selection rules to use when building representations
        :type sel_rules:
        :return:
        :rtype:
        """

        # determine how many states aren't potentially coupled by the operator
        # & then determine which of those are non-orthogonal
        nstates = len(states)

        if self.skipped_indices is not None and tuple(inds) in self.skipped_indices:
            return sp.csr_matrix((1, nstates), dtype='float')

        if check_orthogonality:
            # TODO: selection rules are actually _cheaper_ to apply than this in general, esp. if we focus
            #       only on the number of quanta that can change within the set of indices
            #       so we should support applying them first & then only doing this for the rest
            if use_sel_rule_filtering and sel_rules is not None:
                _, idx = np.unique(inds, return_index=True)
                uinds = inds[np.sort(idx)]
                states, non_orthog = states.apply_sel_rules_along(sel_rules, uinds)
                if len(non_orthog) > 0:
                    states, non_orthog_2 = states.apply_non_orthogonality(inds)#, max_inds=self.fdim)
                    non_orthog = non_orthog[non_orthog_2]
            elif use_sel_sum_filtering and sel_rules is not None:
                states, non_orthog = states.apply_sel_sums(sel_rules, inds)
                if len(non_orthog) > 0:
                    states, non_orthog_2 = states.apply_non_orthogonality(inds)  # , max_inds=self.fdim)
                    non_orthog = non_orthog[non_orthog_2]
            else:
                states, non_orthog = states.apply_non_orthogonality(inds)#, max_inds=self.fdim)
        else:
            non_orthog = np.arange(nstates)


        def return_empty():
            return sp.csr_matrix((1, nstates), dtype='float')

        # if none of the states are non-orthogonal...just don't calculate anything
        if len(non_orthog) == 0:
            return return_empty()
        else:
            if isinstance(funcs, (list, tuple)):
                chunk, all_sels = self._mat_prod_operator_terms(inds, funcs, states, sel_rules)
            else:
                chunk, all_sels = self._direct_prod_operator_terms(inds, funcs, states)

            if all_sels is not None:
                if not (isinstance(all_sels, np.ndarray) and all_sels.dtype == np.dtype(bool)):
                    raise ValueError("bad selection rule application result {}".format(all_sels))
                # logger.log_print('sels: {ns} {sels}', ns=len(all_sels), sels=np.sum(np.where(all_sels)))
                non_orthog = non_orthog[all_sels,]

            if chunk is None:
                return return_empty()

            # if check_orthogonality:
            # finally we make sure that everything we're working with is
            # non-zero because it'll buy us time on dot products later
            non_zero = np.where(np.abs(chunk) >= self.zero_threshold)[0]
            if len(non_zero) == 0:
                return return_empty()
            chunk = chunk[non_zero,]
            non_orthog = non_orthog[non_zero,]

            wat = sp.csr_matrix(
                (
                    chunk,
                    (
                        np.zeros(len(non_zero), dtype='int8'),
                        non_orthog
                    )
                ), shape=(1, nstates))
            # else:
            #     wat = chunk

            return wat

    def _get_pop_sequential(self, inds, idx, check_orthogonality=True, save_to_disk=False):
        """
        Sequential method for getting elements of our product operator tensors

        :param inds:
        :type inds:
        :param idx:
        :type idx:
        :param save_to_disk: whether to save to disk or not; used by the parallelizer
        :type save_to_disk: bool
        :return:
        :rtype:
        """

        res = np.apply_along_axis(self._calculate_single_pop_elements,
                                  -1, inds, self.funcs, idx,
                                  self.selection_rules,
                                  check_orthogonality
                                  )
        if save_to_disk:
            flattened = [x for y in res.flatten() for x in y]
            res = []
            try:
                for array in flattened:
                    with tf.NamedTemporaryFile() as tmp:
                        dump = tmp.name + ".npz"
                    # print(res)
                    # os.remove(tmp.name)
                    sp.save_npz(dump, array, compressed=False)
                    res.append(dump)
            except:
                for file in res:
                    try:
                        os.remove(file)
                    except:
                        pass
                raise

        return res

    @Parallelizer.main_restricted
    def _load_arrays(self, res, parallelizer=None):
        if isinstance(res[0], str):
            res = sum(res, [])

            start = time.time()
            try:
                sparrays = [sp.load_npz(f) for f in res]
            finally: # gotta clean up after myself
                try:
                    for f in res:
                        os.remove(f)
                except Exception as e:
                    # print(e)
                    pass
            self.logger.log_print("reloading arrays from disk: took {e:.3f}s", e=(time.time() - start))
            res = np.array(sparrays, dtype=object)

        return res

    def _get_pop_parallel(self, inds, idx, parallelizer, check_orthogonality=True):
        """
        :param inds:
        :type inds:
        :param idx:
        :type idx:
        :param parallelizer:
        :type parallelizer: Parallelizer
        :return:
        :rtype:
        """

        inds = parallelizer.scatter(inds)
        parallelizer.print("evaluating over inds of shape {s}", s=inds.shape, log_level=parallelizer.logger.LogLevel.Debug)
        start = time.time()
        dump = self._get_pop_sequential(inds, idx, save_to_disk=False, check_orthogonality=check_orthogonality)
        end = time.time()
        parallelizer.print("took {t:.3f}s", t=end-start, log_level=parallelizer.logger.LogLevel.Debug)
        res = parallelizer.gather(dump)
        parallelizer.print("gather took {t:.3f}s", t=time.time()-end, log_level=parallelizer.logger.LogLevel.Debug)
        if isinstance(res, np.ndarray):
            parallelizer.print("got res of shape {s}", s=res.shape, log_level=parallelizer.logger.LogLevel.Debug)
        if parallelizer.on_main:
            res = self._load_arrays(res)

        return res

    @Parallelizer.main_restricted
    # @profile
    def _main_get_elements(self, inds, idx,
                           parallelizer=None,
                           check_orthogonality=True,
                           ):
        """
        Implementation of get_elements to be run on the main process...

        :param idx:
        :type idx:
        :param parallelizer:
        :type parallelizer:
        :return:
        :rtype:
        """

        # we expect this to be an iterable object that looks like
        # num_modes X [bra, ket] X quanta
        try:
            idx = idx.obj # shared proxy deref
        except AttributeError:
            pass

        if not isinstance(idx, BraKetSpace):
            self.logger.log_print("constructing BraKet space")
            idx = BraKetSpace.from_indices(idx, quanta=self.quanta)

        if inds is None: # just a number
            self.logger.log_print("evaluating identity tensor over {} elements".format(len(idx)))
            new = self._get_eye_tensor(idx)
        else:
            mapped_inds, inv_dat = self.filter_symmetric_indices(inds)
            if parallelizer is not None and not isinstance(parallelizer, SerialNonParallelizer):
                self.logger.log_print("evaluating {nel} elements over {nind} unique indices using {cores} processes".format(
                    nel=len(idx),
                    nind=len(mapped_inds),
                    cores=parallelizer.nproc
                ))
                res = self._get_pop_parallel(mapped_inds, idx, parallelizer, check_orthogonality=check_orthogonality)
            else:
                self.logger.log_print("evaluating {nel} elements over {nind} unique indices sequentially".format(
                    nel = len(idx),
                    nind = len(mapped_inds)
                ))
                res = self._get_pop_sequential(mapped_inds, idx, check_orthogonality=check_orthogonality)

            if inv_dat is not None:
                inverse, preinv = inv_dat
                res = res.flatten()[preinv][inverse]

            wat = [x for y in res for x in y]
            new = sp.vstack(wat)

        shp = inds.shape if inds is not None else ()
        res = SparseArray.from_data(new, cache_block_data=False)
        res = res.reshape(shp[:-1] + res.shape[-1:])

        return res

    @Parallelizer.worker_restricted
    def _worker_get_elements(self, idx, parallelizer=None, check_orthogonality=True):
        """

        :param idx:
        :type idx:
        :param parallelizer:
        :type parallelizer:
        :return:
        :rtype:
        """
        try:
            idx = idx.obj # shared proxy deref
        except AttributeError:
            pass
        # worker threads only need to do a small portion of the work
        # and actually inherit most of their info from the parent process
        self._get_pop_parallel(None, idx, parallelizer, check_orthogonality=True)

    def _get_elements(self, inds, idx, full_inds=None, parallelizer=None, check_orthogonality=True):
        """
        Runs the regular element getting algorithm in parallel

        :param idx:
        :type idx:
        :param parallelizer:
        :type parallelizer:
        :return:
        :rtype:
        """
        if full_inds is not None:
            inds = full_inds
        self._worker_get_elements(idx, parallelizer=parallelizer, check_orthogonality=check_orthogonality)
        return self._main_get_elements(inds, idx, parallelizer=parallelizer, check_orthogonality=check_orthogonality)

    def _split_idx(self, idx):
        if self.chunk_size is not None:
            if isinstance(idx, BraKetSpace):
                if len(idx) > self.chunk_size:
                    idx_splits = idx.split(self.chunk_size)
                else:
                    idx_splits = [idx]
            else:
                idx_splits = [idx]
        else:
            idx_splits = [idx]
        return idx_splits

    # from memory_profiler import profile
    # @profile
    def _eval_chunks(self, inds, idx_splits,
                     parallelizer=None,
                     check_orthogonality=True,
                     memory_constrained=False
                     ):

        # import tracemalloc
        #
        # tracemalloc.start()

        chunks = []

        if parallelizer is None:
            parallelizer = self.parallelizer
        if parallelizer is not None and not isinstance(parallelizer, SerialNonParallelizer):
            with parallelizer:

                for n, idx in enumerate(idx_splits):
                    idx: BraKetSpace
                    # parallelizer.printer = self.logger.log_print
                    max_nproc = len(inds.flatten())
                    # idx.load_space_diffs()
                    idx = parallelizer.share(idx)
                    elem_chunk = parallelizer.run(self._get_elements, None, idx,
                                                  check_orthogonality=check_orthogonality,
                                                  main_kwargs={'full_inds': inds},
                                                  comm=None if max_nproc >= parallelizer.nprocs else list(range(max_nproc))
                                                  )
                    if memory_constrained:
                        idx = idx.obj
                        if len(idx_splits) > 1:
                            idx.free()
                        else:
                            idx.clear_cache()
                        idx_splits[n] = None
                        del idx
                        gc.collect()

                    chunks.append(elem_chunk)
        else:

            for n, idx in enumerate(idx_splits):
                idx: BraKetSpace
                elem_chunk = self._main_get_elements(inds, idx,
                                                     parallelizer=None,
                                                     check_orthogonality=check_orthogonality
                                                     )
                if memory_constrained:
                    # self.logger.log_print("mem usage post\n{m}", m="\n".join(str(x) for x in tracemalloc.take_snapshot().statistics('lineno')))
                    if len(idx_splits) > 1:
                        idx.free()
                    else:
                        idx.clear_cache()
                    idx_splits[n] = None
                    del idx
                    gc.collect()

                    # self.logger.log_print("mem usage freed\n{m}", m="\n".join(str(x) for x in tracemalloc.take_snapshot().statistics('lineno')))

                chunks.append(elem_chunk)

        return chunks

    # @profile
    def _construct_elem_array(self, chunks):

        if all(isinstance(x, np.ndarray) for x in chunks):
            elems = np.concatenate(chunks, axis=0)
        else:
            elems = chunks[0].concatenate(*chunks[1:])

        return elems

    # @profile
    def get_elements(self, idx,
                     parallelizer=None,
                     check_orthogonality=True,
                     memory_constrained=False
                     # need to work in further cache clearing to account for mem. constraints
                     ):
        """
        Calculates a subset of elements

        :param idx: bra and ket states as tuples of elements
        :type idx: BraKetSpace
        :return:
        :rtype:
        """

        idx_splits = self._split_idx(idx)

        inds = self.get_inner_indices()
        if inds is None:
            return self._get_elements(inds, idx, check_orthogonality=check_orthogonality)

        chunks = self._eval_chunks(inds, idx_splits,
                                   parallelizer=parallelizer,
                                   check_orthogonality=check_orthogonality,
                                   memory_constrained=memory_constrained
                                   )

        arr = self._construct_elem_array(chunks)
        return arr

    @staticmethod
    def _get_dim_string(dims):
        if len(dims) < 7:
            return ", ".join(str(s) for s in dims)
        else:
            return "{}, ({} more), {}".format(
                ", ".join(str(s) for s in dims[:2]),
                len(dims) - 4,
                ", ".join(str(s) for s in dims[-2:])
            )
    def __repr__(self):
        return "{}(<{}>, {})".format(
            type(self).__name__,
            self._get_dim_string(self.shape),
            self.funcs
        )

    @classmethod
    def _get_skip_trans(cls, inds, ndim):
        # TODO: this needs generalization to work properly with different
        #       forms of selection rules
        if inds is None:
            return None

        skips = set()
        for i in inds:
            # get every possible transition phase
            for phase in itertools.product(*[[-1, 1] for i in range(len(i))]):
                base = [0] * ndim
                for j, p in zip(i, phase):
                    base[j] += p
                skips.add(tuple(base))
        return np.array(list(skips))

    def get_transformed_space(self, base_space, rules=None, parallelizer=None, logger=None, **opts):
        """
        Returns the space one would get from applying
        the selection rules from this operator

        :param base_space:
        :type base_space: BasisStateSpace
        :return:
        :rtype: SelectionRuleStateSpace
        """
        if rules is None:
            rules = self.selection_rules
        if parallelizer is None:
            parallelizer = self.parallelizer
        new = base_space.apply_selection_rules(rules,
                                                parallelizer=parallelizer,
                                                logger=logger,
                                                **opts
                                                )
        skips = self._get_skip_trans(self.skipped_indices, base_space.ndim)
        if skips is not None:
            new = new.filter_transitions(skips)
        return new

    def _calculate_single_transf(self, inds, funcs, base_space, sel_rules):
        """
        Calculates terms for a single product operator.
        Assumes orthogonal product bases.

        :param inds: the index in the total operator tensor (i.e. which of the funcs to use)
        :type inds:
        :param funcs: the functions to use when generating representations, must return matrices
        :type funcs:
        :param states: the states to compute terms between stored like ((s_1l, s_1r), (s_2l, s_2r), ...)
                        where `s_il` is the set of quanta for mode i for the bras and
                        `s_ir` is the set of quanta for mode i for the kets
        :type states: BraKetSpace
        :param quants: the total quanta to use when building representations
        :type quants:
        :param sel_rules: the selection rules to use when building representations
        :type sel_rules:
        :return:
        :rtype:
        """

        # once we're doing this over all possible indices I think I've lost any benefit

        # TODO: edit direct sum generator to be a true generator where I can directly resume
        #       the calculation over a different set of target dimensions if I need to or something
        #       although I'm low-key not sure that's feasible
        #       _Alternately_ allow for splitting by which indices were touched rather than what the
        #       input state was

        # trying to not calculate anything unnecessary
        uinds = np.unique(inds)
        sr = [r for r in sel_rules if len(r) == len(uinds) or (len(uinds) == 1 and len(r) == 0)]
        states = base_space.apply_selection_rules(sr, target_dimensions=uinds)
        brakets = states.get_representation_brakets()
        vals = self._calculate_single_pop_elements(inds, funcs, brakets, sel_rules, check_orthogonality=False)

        return vals, brakets#, states
    def _get_transf_sequential(self, inds, base_space):#, save_to_disk=False):
        """
        Sequential method for getting elements of our product operator tensors

        :param inds:
        :type inds:
        :param base_space:
        :type base_space:
        :param save_to_disk: whether to save to disk or not; used by the parallelizer
        :type save_to_disk: bool
        :return:
        :rtype:
        """

        res = np.apply_along_axis(self._calculate_single_transf,
                                  -1, inds, self.funcs, base_space, self.selection_rules
                                  )

        # spaces = [r[2] for r in res]
        brakets = [r[1] for r in res]
        res = [r[0] for r in res]

        # total_space = res[0][2]
        # total_brakets = res[0][1]
        # for s in res[1:]:
        #     total_brakets = total_brakets.concatenate(s[1]) # to preserve ordering since union will destroy that
        #     total_space = total_space.union(s[2])
        # res = np.concatenate([r[0] for r in res])

        # if save_to_disk:
        #     flattened = [x for y in res.flatten() for x in y]
        #     res = []
        #     try:
        #         for array in flattened:
        #             with tf.NamedTemporaryFile() as tmp:
        #                 dump = tmp.name + ".npz"
        #             # print(res)
        #             # os.remove(tmp.name)
        #             sp.save_npz(dump, array, compressed=False)
        #             res.append(dump)
        #     except:
        #         for file in res:
        #             try:
        #                 os.remove(file)
        #             except:
        #                 pass
        #         raise
        #
        # raise Exception(res, total_space, total_brakets, total_brakets.bras.excitations[5], inds, total_brakets.kets.excitations[5])

        return res, brakets
    @Parallelizer.main_restricted
    def _apply_transformations_sequential(self, inds, base_space, perm_class_map, parallelizer=None):
        """
        Implementation of get_elements to be run on the main process...

        :param idx:
        :type idx:
        :param parallelizer:
        :type parallelizer:
        :return:
        :rtype:
        """

        raise NotImplementedError("no")

        if inds is None:  # just a number
            self.logger.log_print("returning identity tensor")
            new = self._get_eye_tensor(base_space)
        else:
            mapped_inds, inverse = self.filter_symmetric_indices(inds) # I can't actually reduce here...
            inds = np.reshape(inds, (-1, inds.shape[-1]))
            if parallelizer is not None and not isinstance(parallelizer, SerialNonParallelizer):
                raise NotImplementedError("still working up parallelism")
                self.logger.log_print(
                    "evaluating {nel} elements over {nind} unique indices using {cores} processes".format(
                        nel=len(idx),
                        nind=len(mapped_inds),
                        cores=parallelizer.nproc
                    ))
                res = self._get_pop_parallel(mapped_inds, idx, parallelizer)
            else:
                self.logger.log_print("transforming space of size {nel} over {nind} unique indices sequentially".format(
                    nel=len(base_space),
                    nind=len(mapped_inds)
                ))

                res, brakets = self._get_transf_sequential(mapped_inds, base_space)

            if inverse is not None: # fix the reductions by symmetry
                res = [res[i] for i in inverse]
                brakets = [brakets[i] for i in inverse]

            # TODO: this will need to be _a lot_ faster if this is going to become the default
            #       method for getting representations...
            # now we need to apply the permutations which
            # we'll do in a slightly dumb/inefficient way
            # by determining which brakets correspond to which
            # input classes and using this to do the necessary
            # permutations of all relevant components
            full_res = []
            full_bras = []
            full_kets = []
            full_inds = []
            for r, b, i in zip(res, brakets, inds):
                res_groups, sorting = nput.group_by(r, b.bras.excitations)
                ket_groups, _ = nput.group_by(b.kets.excitations, b.bras.excitations, sorting=sorting)
                res_keys, res_vals = res_groups
                # raise Exception(res_groups[0], ket_groups[0])
                for key, vals, kets in zip(res_keys, res_vals, ket_groups[1]):
                    for test_key, perms, inv_perms in perm_class_map:
                        if np.all(key == test_key):
                            # permute kets
                            perm_kets = np.array([k[perms] for k in kets]).transpose((1, 0, 2))
                            perm_bras = np.broadcast_to(key[perms][:, np.newaxis, :], (len(perms), len(vals), key.shape[-1]))
                            perm_inds = np.broadcast_to(inv_perms[:, np.newaxis, i], (len(perms), len(vals), len(i)))
                            perm_vals = np.broadcast_to(vals[np.newaxis], (len(perms), len(vals)))
                            full_res.append(perm_vals.reshape(-1))
                            full_bras.append(perm_bras.reshape(-1, perm_bras.shape[-1]))
                            full_kets.append(perm_kets.reshape(-1, perm_kets.shape[-1]))
                            full_inds.append(perm_inds.reshape(-1, perm_inds.shape[-1]))
                            break
                    else:
                        raise ValueError("couldn't find key {} in input keys?".format(key))

            full_bras = np.concatenate(full_bras, axis=0)
            full_kets = np.concatenate(full_kets, axis=0)
            full_inds = np.concatenate(full_inds)
            full_res = np.concatenate(full_res)

            new_brakets = BraKetSpace(
                BasisStateSpace(base_space.basis, full_bras, mode=BasisStateSpace.StateSpaceSpec.Excitations),
                BasisStateSpace(base_space.basis, full_kets, mode=BasisStateSpace.StateSpaceSpec.Excitations),
            )
            _, _, bk_inds = nput.unique(np.concatenate([full_bras, full_kets], axis=1), return_inverse=True)
            real_full_inds = np.concatenate([full_inds, bk_inds[:, np.newaxis]], axis=1).T

            res = SparseArray.from_data(
                (
                    full_res,
                    real_full_inds
                ),
                shape=(self.mode_n, self.mode_n, np.max(bk_inds)+1)
            )

            # raise Exception("oookay")
            #
            # res = np.concatenate(res)
            # full_brakets = brakets[0]
            # for s in brakets[1:]:
            #     full_brakets = full_brakets.concatenate(s)

        return res, new_brakets

    def apply_reduced(self, base_space, parallelizer=None, logger=None):
        """
        Takes a base space as input and applies the held selection rules in semi-efficient
        fashion only on the indices that can change and then uses this to compute all matrix
        elements, returning then the final generated space

        :param base_space:
        :type base_space: BasisStateSpace | PermutationallyReducedStateSpace
        :return:
        :rtype: tuple[SparseArray, BraKetSpace]
        """

        if not isinstance(base_space, PermutationallyReducedStateSpace):
            base_space = base_space.permutationally_reduce()

        rep_space = base_space.representative_space()
        if self.chunk_size is not None and len(rep_space) > self.chunk_size:
            raise NotImplementedError("not handling chunking yet...")
            idx_splits = rep_space.split(self.chunk_size)
        else:
            idx_splits = [rep_space]

        inds = self.get_inner_indices(reduced_inds=False)
        perm_class_map = [(e, p, np.argsort(p, axis=1)) for e,p in zip(rep_space.excitations, base_space.equivalence_classes)]
        chunks = []
        brakets = []
        for idx in idx_splits:
            if inds is None:
                return self._apply_transformations_sequential(inds, idx)

            if parallelizer is None:
                parallelizer = self.parallelizer

            if parallelizer is not None and not isinstance(parallelizer, SerialNonParallelizer):
                raise NotImplementedError("Don't have parallelism yet")
                parallelizer.printer = self.logger.log_print
                elem_chunk = parallelizer.run(self._get_elements, None, idx,
                                              main_kwargs={'full_inds':inds},
                                              comm=None if len(inds) >= parallelizer.nprocs else list(range(len(inds)))
                                              )
            else:
                elem_chunk, elem_brakets = self._apply_transformations_sequential(inds, idx, perm_class_map, parallelizer=None)

            brakets.append(elem_brakets)
            chunks.append(elem_chunk)

        brakets = brakets[0]
        chunks = chunks[0]

        # raise Exception(chunks, brakets)

        # if all(isinstance(x, np.ndarray) for x in chunks):
        #     elems = np.concatenate(chunks, axis=0)
        # else:
        #     from functools import reduce
        #     elems = reduce(lambda a, b: a.concatenate(b), chunks[1:], chunks[0])

        return chunks, brakets

class ContractedOperator(Operator):
    """
    Provides support for terms that look like `pGp` or `p(dG/dQ)Qp` by
    expanding them out as the pure operator component that depends on the basis states (i.e. `pp` or `pQp`)
    and doing the appropriate tensor contractions with the expansion coefficients (i.e. `G` or `dG/dQ`)
    """

    def __init__(self, coeffs, funcs, quanta,
                 prod_dim=None, axes=None, symmetries=None,
                 selection_rules=None,
                 selection_rule_steps=None,
                 zero_threshold=1.0e-14,
                 skipped_indices=None,
                 skipped_coefficient_threshold=None,
                 chunk_size=None,
                 parallelizer=None,
                 logger=None
                 ):
        """

        :param coeffs: The tensor of coefficients contract with the operator representation (`0` means no term)
        :type coeffs: np.ndarray | int
        :param funcs: The functions use to calculate representation
        :type funcs: callable | Iterable[callable]
        :param quanta: The number of quanta to do the deepest-level calculations up to
        :type quanta: int | Iterable[int]
        :param axes: The axes to use when doing the contractions
        :type axes: Iterable[int] | None
        :param symmetries: The symmetries to pass through to `Operator`
        :type symmetries: Iterable[int] | None
        :param prod_dim:
        :type prod_dim:
        :param selection_rules:
        :type selection_rules:
        :param parallelizer:
        :type parallelizer:
        :param logger:
        :type logger:
        :param zero_threshold:
        :type zero_threshold:
        :param chunk_size: number of elements that can be evaluated at once (for memory reasons)
        :type chunk_size: int | None
        """
        self.coeffs = coeffs
        self.axes = axes
        if (
                skipped_indices is None
                and skipped_coefficient_threshold is not None
                and not (isinstance(self.coeffs, (int, float, np.integer, np.floating)) and self.coeffs == 0)
        ):
            skipped_indices = np.where(np.abs(coeffs) < skipped_coefficient_threshold)
            if len(skipped_indices) > 0 and len(skipped_indices[0]) > 0:
                # print("!", len(skipped_indices), self.coeffs.ndim, skipped_indices)
                # if len(skipped_indices) > 5:
                #     raise Exception(skipped_indices)
                skipped_indices = np.array(skipped_indices).T
            else:
                skipped_indices = None
        super().__init__(funcs, quanta, symmetries=symmetries, prod_dim=prod_dim,
                         selection_rules=selection_rules,
                         selection_rule_steps=selection_rule_steps,
                         zero_threshold=zero_threshold,
                         parallelizer=parallelizer,
                         logger=logger,
                         skipped_indices=skipped_indices
                         )
        self.chunk_size = chunk_size

    @property
    def is_zero(self):
        return (isinstance(self.coeffs, (int, float, np.integer, np.floating)) and self.coeffs == 0)

    def _get_element_block(self, idx, parallelizer=None, check_orthogonality=True, memory_constrained=False):
        c = self.coeffs
        if not nput.is_numeric(c):
            # takes an (e.g.) 5-dimensional SparseTensor and turns it into a contracted 2D one
            axes = self.axes
            if axes is None:
                axes = (tuple(range(c.ndim)),) * 2
            subTensor = super().get_elements(idx, parallelizer=parallelizer, check_orthogonality=check_orthogonality, memory_constrained=memory_constrained)

            # we collect here to minimize the effect of memory spikes if possible
            # self.clear_cache()
            # SparseArray.clear_cache()
            gc.collect()
            if isinstance(subTensor, np.ndarray):
                if len(axes[1]) > 0:
                    contracted = np.tensordot(subTensor.squeeze(), c, axes=axes)
                else:
                    # TODO: make this broadcasting more robust
                    contracted = c[np.newaxis, :] * subTensor.squeeze()[:, np.newaxis]
            else:
                with subTensor.cache_options(enabled=False):
                    if len(axes[1]) > 0:
                        contracted = subTensor.tensordot(c, axes=axes).squeeze()
                    else:
                        # TODO: make this broadcasting more robust
                        subTensor = subTensor.expand_dims(-1)
                        contracted = subTensor * c[np.newaxis, :]

            # if self.fdim > 3:
            #     raise RuntimeError("wwwwooooof")

        elif c == 0:
            contracted = 0  # a short-circuit
        else:
            subTensor = super().get_elements(idx, check_orthogonality=check_orthogonality)
            if c == 1:
                return subTensor
            contracted = c * subTensor

        gc.collect()

        return contracted

    def get_elements(self, idx, parallelizer=None, check_orthogonality=True, memory_constrained=False):
        """
        Computes the operator values over the specified indices

        :param idx: which elements of H0 to compute
        :type idx: Iterable[int]
        :return:
        :rtype:
        """
        c = self.coeffs
        if isinstance(c, (int, np.integer, float, np.floating)) and c == 0:
            return 0

        idx_splits = self._split_idx(idx)

        # need a special case for diagonal tensors
        chunks = [self._get_element_block(idx, check_orthogonality=check_orthogonality, memory_constrained=memory_constrained) for idx in idx_splits]
        if all(isinstance(x, np.ndarray) for x in chunks):
            subchunks = [np.array([y]) if y.shape == () else y for y in chunks]
            contracted = np.concatenate(subchunks, axis=0)
        else:
            contracted = chunks[0].concatenate(*chunks[1:])
            # print([x.shape for x in chunks])
            # print(contracted.shape)
            #
            # raise Exception(contracted.asarray() -
            #                 np.concatenate([x.asarray() for x in chunks], axis=0)
            #                 )
            # assert np.all(contracted == np.concatenate([x.asarray() for x in chunks], axis=0))
        # print(contracted.shape)
        return contracted

    def apply_reduced(self, base_space, parallelizer=None, logger=None):

        c = self.coeffs
        if isinstance(c, (int, np.integer, float, np.floating)) and c == 0:
            return 0

        c = self.coeffs
        if not isinstance(c, (int, np.integer, float, np.floating)):
            # takes an (e.g.) 5-dimensional SparseTensor and turns it into a contracted 2D one
            axes = self.axes
            if axes is None:
                axes = (tuple(range(c.ndim)),) * 2

            subTensor, brakets = super().apply_reduced(base_space, parallelizer=parallelizer, logger=logger)

            # we collect here to minimize the effect of memory spikes if possible
            # self.clear_cache()
            # SparseArray.clear_cache()
            gc.collect()
            if isinstance(subTensor, np.ndarray):
                contracted = np.tensordot(subTensor.squeeze(), c, axes=axes)
            else:
                contracted = subTensor.tensordot(c, axes=axes).squeeze()

            # if self.fdim > 3:
            #     raise RuntimeError("wwwwooooof")

        elif c == 0:
            contracted = 0  # a short-circuit
            brakets = None
        else:
            subTensor, brakets = super().apply_reduced(base_space, parallelizer=parallelizer, logger=logger)
            if c == 1:
                return subTensor
            contracted = c * subTensor

        gc.collect()

        return contracted, brakets

        # return chunks, brakets


    def __repr__(self):
        return "{}(<{}>x<{}>, {})".format(
            type(self).__name__,
            None if not hasattr(self.coeffs, 'shape') else self._get_dim_string(self.coeffs.shape),
            self._get_dim_string(self.shape),
            self.funcs
        )




"""
Provides core Data-related types and data structures.
Intended to be high-level data as opposed to the lower-level stuff in `McUtils.Data`.
That means including stuff like dipole and potential energy surfaces that know how to compute their own properties.
We also have expressions for G-matrix elements from Frederick and Woywood to use with `sympy`.
"""


__all__ = []
from .Surfaces import *; from .Surfaces import __all__ as exposed
__all__ += exposed
from .KEData import *; from .KEData import __all__ as exposed
__all__ += exposed


"""
Provides a general base spectrum class that can be extended to new fancy spectral forms
"""

import numpy as np, json, abc
import McUtils.Plots as plt
from McUtils.Data import UnitsData

__all__ = [
    "BaseSpectrum",
    "DiscreteSpectrum",
    "ContinuousSpectrum",
    "BroadenedSpectrum",
    "MultiSpectrum"
]

# TODO: should add in support for various flavors of unit conversion
#       like oscillator strengths to km mol^-1 units and whatnot
class BaseSpectrum:
    """
    Base class to support spectral operation
    """
    def __init__(self, frequencies, intensities, **meta):
        """
        :param frequencies: frequency list
        :type frequencies: np.ndarray
        :param intensities: intensity list
        :type intensities: np.ndarray
        :param meta: metadata
        :type meta:
        """
        self.frequencies = frequencies
        self.intensities = intensities
        self.meta = meta

    def take_subspectrum(self, pos):
        """
        Takes a subset of frequencies/intensities specified by `pos`

        :param pos:
        :type pos:
        :return:
        :rtype:
        """

        return type(self)(self.frequencies[pos], self.intensities[pos], **self.meta)

    def __getitem__(self, item):

        f = self.frequencies[item]
        if isinstance(f, (int, float, np.integer, np.floating)):
            return (f, self.intensities[item])
        else:
            return self.take_subspectrum(item)

    def frequency_filter(self, freq_min, freq_max):
        """
        Filters by frequencies >= `freq_min` and <= `freq_max`

        :param freq_min: min frequency
        :type freq_min: float
        :param freq_max: max frequency
        :type freq_max: float
        :return: subspectrum
        :rtype: BaseSpectrum
        """

        freqs = self.frequencies
        return self.take_subspectrum(np.where(np.logical_and(freq_min <= freqs, freqs <= freq_max))[0])

    def intensity_filter(self, int_min, int_max):
        """
        Filters by intensities >= `int_min` and <= `int_max`

        :param int_min: min intensity
        :type int_min: float
        :param int_max: max intensity
        :type int_max: float
        :return: subspectrum
        :rtype: BaseSpectrum
        """

        ints = self.intensities
        return self.take_subspectrum(np.where(np.logical_and(int_min <= ints, ints <= int_max))[0])


    def save(self, file):
        """
        Saves the spectrum in JSON format

        :param file: str | file-like object
        :type file:
        :return:
        :rtype:
        """

        dump = json.dumps({
            "frequencies": self.frequencies,
            "intensities": self.intensities,
            "metadata": self.meta
        })
        if isinstance(file, str):
            with open(file, "w") as f:
                f.write(dump)
        else:
            file.write(dump)

    @classmethod
    def load(cls, file):
        """
        Saves a spectrum from a JSON file

        :param file: str | file-like object
        :type file:
        :return:
        :rtype:
        """

        if isinstance(file, str):
            with open(file, 'r') as f:
                dump = json.load(f)
        else:
            dump = json.load(file)

        return cls(dump['frequencies'], dump['intensities'], **dump['metadata'])

    @abc.abstractmethod
    def plot(self, **opts):
        """
        A stub so that subclasses can implement their own `plot` methods

        :param opts: plotting options to be fed through to whatever the plotting function uses
        :type opts:
        :return:
        :rtype:
        """
        raise NotImplementedError("{}: unclear how to plot spectrum".format(
            type(self).__name__
        ))


class DiscreteSpectrum(BaseSpectrum):
    """
    Concrete implementation of `BaseSpectrum` that exists
    solely to allow for plotting and broadening.
    """

    @classmethod
    def from_transition_moments(cls, frequencies, transition_moments, **meta):
        """
        Assumes frequencies and transition moments in a.u.

        :param frequencies:
        :type frequencies:
        :param transition_moments:
        :type transition_moments:
        :return:
        :rtype:
        """

        oscillator_strengths = np.linalg.norm(transition_moments, axis=-1) ** 2
        units = UnitsData.convert("OscillatorStrength", "KilometersPerMole")
        h2w = UnitsData.convert("Hartrees", "Wavenumbers")
        return cls(
            frequencies * h2w,
            units * frequencies * oscillator_strengths,
            **meta
        )

    def normalize(self, which=None):
        return type(self)(
            self.frequencies,
            self.intensities / (np.max(self.intensities) if which is None else self.intensities[which]),
            **self.meta
        )

    def plot(self,
             figure=None,
             plot_style=None,
             **opts
             ):
        """
        Plots a spectrum using `McUtils.Plots.StickSpectrum`

        :param figure: figure to plot the spectrum on
        :type figure: None | McUtils.Plots.Graphics
        :param opts: any of the many, many options supported by `McUtils.Plots.Graphics`
        :type opts:
        :return:
        :rtype:
        """

        # suuuuper straightforward
        return plt.StickPlot(self.frequencies, self.intensities,
                             figure=figure,
                             plot_style=plot_style,
                             **opts
                             )

    def broaden(self, broadening_type="gaussian", breadth=10):
        """
        Applies a broadening to the spectrum

        :param broadening_type:
        :type broadening_type:
        :param breadth:
        :type breadth:
        :return:
        :rtype:
        """
        return BroadenedSpectrum(self.frequencies, self.intensities,
                                 broadening_type=broadening_type, breadth=breadth,
                                 **self.meta
                                 )

class ContinuousSpectrum(BaseSpectrum):
    """
    Concrete implementation of `BaseSpectrum` that exists
    solely to allow for plotting & maybe some day for interchange with like experimental formats
    """

    def plot(self,
             figure=None,
             filled=False,
             plot_style=None,
             **opts
             ):
        """
        Plots a spectrum using `McUtils.Plots.Plot`

        :param figure: figure to plot the spectrum on
        :type figure: None | McUtils.Plots.Graphics
        :param opts: any of the many, many options supported by `McUtils.Plots.Graphics`
        :type opts:
        :return:
        :rtype:
        """

        # suuuuper straightforward
        main = plt.Plot(self.frequencies, self.intensities,
                         figure=figure,
                         plot_style=plot_style,
                         **opts
                         )
        if filled:
            plt.FilledPlot(self.frequencies, self.intensities,
                     figure=main,
                     plot_style=plot_style,
                     **opts
                     )
        return main

class BroadenedSpectrum(BaseSpectrum):
    """
    A stick spectrum with associated broadening function
    """

    def __init__(self, frequencies, intensities, broadening_type="gaussian", breadth=10, **meta):
        """
        :param frequencies:
        :type frequencies:
        :param intensities:
        :type intensities:
        :param broadening_type: the type of broadening to apply (can be any function)
        :type broadening_type: "gaussian" | "lorentzian" | function
        :param breadth: the breadth or list of breads for the peaks in the spectrum
        :type breadth:
        :param meta:
        :type meta:
        """

        super().__init__(frequencies, intensities, **meta)

        self.broadening_type = broadening_type
        self.breadth = breadth

    def _eval_gauss_broadening(self, pts, height, center, breadth, target_zero=.1, adjust_width=True):
        """
        Evaluates a Gaussian centered around `center` with breadth `breadth` at `pts`

        :param pts:
        :type pts: np.ndarray
        :param center:
        :type center: float
        :param breadth:
        :type breadth: float
        :return:
        :rtype:
        """
        from scipy.stats import norm

        if adjust_width:
            h = height
            if h < 1:
                h = 1
            z = target_zero / h
            breadth = np.sqrt(breadth**2/(2*np.log(1/z))) # chosen to make the pdf(center-breadth) == target
        bd = norm(loc=center, scale=breadth)
        return height * bd.pdf(pts) * (np.sqrt(2 * np.pi) * breadth) # remove the normalization

    def _eval_lorentz_broadening(self, pts, height, center, breadth):
        """
        Evaluates a Lorentzian centered around `center` with breadth `breadth` at `pts`

        :param pts:
        :type pts: np.ndarray
        :param center:
        :type center: float
        :param breadth:
        :type breadth: float
        :return:
        :rtype:
        """
        from scipy.stats import cauchy

        bd = cauchy(loc=center, scale=breadth).pdf
        # if height < 1:
        #     height = 1
        return height * bd(pts) * (np.pi * breadth )  # remove the normalization

    def _get_pts(self, step_size=.5, freq_min=None, freq_max=None, stddevs=5, adjust_width=True):
        """
        Evaluates the points needed to plot the broadened spectrum

        :param step_size:
        :type step_size:
        :param freq_min:
        :type freq_min:
        :param freq_max:
        :type freq_max:
        :return:
        :rtype:
        """

        freqs = self.frequencies
        ints = self.intensities

        if isinstance(self.breadth, (int, float, np.integer, np.floating)):
            breadths = [self.breadth] * len(freqs)
        else:
            breadths = self.breadth

        if freq_min is None:
            min_pos = np.argmin(freqs)
            freq_min = freqs[min_pos] - stddevs*breadths[min_pos]
        if freq_max is None:
            max_pos = np.argmax(freqs)
            freq_max = freqs[max_pos] + stddevs*breadths[max_pos]

        bt = self.broadening_type

        if isinstance(bt, str):
            if bt.lower() == 'gaussian':
                bt = self._eval_gauss_broadening
            elif bt.lower() == 'lorentzian':
                bt = self._eval_lorentz_broadening
            else:
                raise ValueError("{}.{}: don't know how to handle broadening type '{}'".format(
                    type(self).__name__,
                    'plot',
                    bt.lower()
                ))

        freq_vals = np.arange(freq_min, freq_max, step_size)
        vals = np.sum([bt(freq_vals, i, c, b, adjust_width=adjust_width) for i, c, b in zip(ints, freqs, breadths)], axis=0)

        return freq_vals, vals

    def plot(self, step_size=.5, freq_min=None, freq_max=None, figure=None, plot_style=None, filled=False, adjust_width=True, **opts):
        """
        Applies the broadening then plots it using `McUtils.Plots.Plot`

        :param step_size: step size to use when getting evaluation points for evaluating the broadening
        :type step_size:
        :param freq_min: min freq for evaluation
        :type freq_min:
        :param freq_max: max freq for evaluation
        :type freq_max:
        :param figure:
        :type figure:
        :param plot_style:
        :type plot_style:
        :param opts:
        :type opts:
        :return:
        :rtype:
        """

        freqs, ints = self._get_pts(step_size=step_size, freq_min=freq_min, freq_max=freq_max, adjust_width=adjust_width)
        if plot_style is None:
            plot_style = {}
        if filled:
            main = plt.FilledPlot(freqs, ints,
                     figure=figure,
                     plot_style=plot_style,
                     **opts
                     )
            plt.Plot(freqs, ints,
                        figure=main,
                        plot_style=plt.Plot.filter_options(plot_style),
                        **plt.Plot.filter_options(opts)
                        )
        else:
            main = plt.Plot(freqs, ints,
                     figure=figure,
                     plot_style=plot_style,
                     **opts
                     )
        return main

class MultiSpectrum:
    """
    A wrapper for multiple spectra, really just for the plotting convenience
    """
    """
        Base class to support spectral operation
        """

    def __init__(self, spectra:"Iterable[BaseSpectrum]", **meta):
        """
        :param frequencies: frequency list
        :type frequencies: np.ndarray
        :param intensities: intensity list
        :type intensities: np.ndarray
        :param meta: metadata
        :type meta:
        """
        self.specs = spectra
        self.meta = meta

    def __getitem__(self, item):

        specs = self.specs[item]
        if isinstance(specs, BaseSpectrum):
            return specs
        else:
            return type(self)(specs)

    def frequency_filter(self, freq_min, freq_max):
        """
        Filters by frequencies >= `freq_min` and <= `freq_max`

        :param freq_min: min frequency
        :type freq_min: float
        :param freq_max: max frequency
        :type freq_max: float
        :return: subspectrum
        :rtype: MultiSpectrum
        """

        return type(self)([
            s.frequency_filter(freq_min, freq_max)
            for s in self.specs
        ])

    def intensity_filter(self, int_min, int_max):
        """
        Filters by intensities >= `int_min` and <= `int_max`

        :param int_min: min intensity
        :type int_min: float
        :param int_max: max intensity
        :type int_max: float
        :return: subspectrum
        :rtype: BaseSpectrum
        """

        return type(self)([
            s.intensity_filter(int_min, int_max)
            for s in self.specs
        ])

    def plot(self, figure=None, **opts):
        """
        A just plots all the spectra on the same figure

        :param opts: plotting options to be fed through to whatever the plotting function uses
        :type opts:
        :return:
        :rtype:
        """

        for p in self.specs:
            subfig = p.plot(figure=figure, **opts)
            if figure is None:
                figure = subfig
        return figure


"""
Wavefun provides a basic framework for working with Wavefunctions that can be subclassed and built upon
"""

from .Wavefunctions import *

__all__ = []
from .Wavefunctions import __all__ as exposed
__all__ += exposed


import collections, functools, numpy as np
# might be worth excising theses since this could be standalone otherwise?
from McUtils.Data import AtomData, UnitsData
from McUtils.Zachary import FiniteDifferenceDerivative, RBFDInterpolator
import McUtils.Numputils as nput

from ..Molecools import Molecule, MolecularZMatrixCoordinateSystem
from ..Molecools.Properties import PropertyManager

__all__ = [
    "AIMDSimulator",
    "PairwisePotential"
]

class AIMDSimulator:
    def __init__(self,
                 atoms,
                 coordinates,
                 force_function,
                 atomic_structures=None,
                 internals=None,
                 velocities=0,
                 track_kinetic_energy=False,
                 track_velocities=False,
                 timestep=.1,
                 sampling_rate=1
                 ):
        self.masses = np.array([
                AtomData[a, "Mass"] * UnitsData.convert("AtomicMassUnits", "ElectronMass")
            if isinstance(a, str) else a for a in atoms
        ])
        coords = np.asanyarray(coordinates)
        if coords.ndim == 1:
            coords = coords[np.newaxis]

        if atomic_structures is None:
            atomic_structures = (
                coords.shape[-1] == 3
                    and
                coords.shape[-2] == len(self.masses)
            )
        self._atomic_structs = atomic_structures
        if self._atomic_structs:
            if coords.ndim == 2: coords = coords[np.newaxis]
            self._mass = np.expand_dims(self.masses, -1)
            for _ in range(coords.ndim - 2):
                self._mass = np.expand_dims(self._mass, 0)
            # self._mass = self.masses[np.newaxis, :,  np.newaxis]
        else: # regular coordinates
            self._mass = self.masses[np.newaxis]
            for _ in range(coords.ndim - 2):
                self._mass = np.expand_dims(self._mass, 0)
        self.coords = coords
        # for _ in range(coordinates.ndim - 2):
        #     self._mass = np.expand_dims(self._mass, -1)
        if isinstance(velocities, (int, float, np.integer, np.floating)):
            velocities = np.full_like(self.coords, velocities)
        velocities = np.asanyarray(velocities)
        self.velocities = velocities
        if internals is not None:
            raise NotImplementedError("haven't fully managed getting embedding right yet...")
            if not isinstance(internals, MolecularZMatrixCoordinateSystem):
                base_mol = Molecule(
                    ['H']*len(self.masses),
                    coords=self.coords,
                    masses=self.masses,
                    internals=internals
                )
                internals = [base_mol.coords.system, base_mol.internal_coordinates.system]
            else:
                internals = [internals.molecule.coords.system, internals]
        self._internals = internals
        self.force_function = force_function
        self._prev_forces = None

        self.trajectory = collections.deque()
        self.trajectory.append(self.coords)
        if track_kinetic_energy:
            self.kinetic_energies = collections.deque()
            self.kinetic_energies.append(
                self._ke(self._mass, self.velocities)
            )
        else:
            self.kinetic_energies = None
        if track_velocities:
            self.velocity_deque = collections.deque()
            self.velocity_deque.append(self.velocities)
        else:
            self.velocity_deque = None
        self.steps = 0
        self.dt = timestep
        self.sampling_rate = sampling_rate

    @classmethod
    def mode_energies_to_velocities(cls, modes, masses, energy_splits, inverse=None):

        masses = np.asanyarray(masses)
        e_part = np.asanyarray(energy_splits)
        smol = e_part.ndim == 1
        if smol: e_part = e_part[np.newaxis]

        # if not np.allclose(np.linalg.norm(modes, axis=0), np.ones(modes.shape[1])):  # i.e. came out of a generalized eigenvalue run
        #     modes = modes.reshape(-1, 3, modes.shape[1]) * np.sqrt(masses[:, np.newaxis, np.newaxis])
        #     modes = modes.reshape(-1, modes.shape[-1])

        if inverse is None:
            if not np.allclose(np.linalg.norm(modes, axis=0), np.ones(modes.shape[1])):
                raise ValueError("modes aren't normalized -> non-unitary transformation")
            inverse = modes.T

        tripmass = np.broadcast_to(masses[:, np.newaxis], (len(masses), 3)).flatten()
        g = inverse @ np.diag(tripmass) @ modes

        gv, Q = np.linalg.eigh(g)
        sorting = np.argsort(np.argmax(Q ** 2, axis=0))
        gv = gv[sorting]  # maximum similarity to OG vectors
        Q = Q[:, sorting]

        axes = (Q.T @ inverse).reshape(modes.shape[1], -1, 3)
        v_part = np.sign(e_part) * np.sqrt(2 * np.abs(e_part)) # / gv[np.newaxis])
        vels = np.sum(
            axes[np.newaxis] *
                v_part[:, :, np.newaxis, np.newaxis],
            axis=1
        )

        return vels

    #TODO: add internal coordinate propagation following the Krimm paper using the B-matrix
    #      or not...maybe I just need a faster Jacobian?
    def get_forces(self, coords):
        if self._internals is not None:
            ccs, ics = self._internals
            icoords = ccs.convert_coords(coords, ics)
            if not isinstance(icoords, np.ndarray):
                icoords = icoords[0] # converter info
            forces = self.force_function(icoords)
            jacs = ccs(coords).jacobian(ics, order=1, all_numerical=True)[0] #dRdX?
            ncs = np.prod(coords.shape[-2:], dtype=int)
            jacs = np.reshape(np.moveaxis(jacs, 1, 0), (len(coords), ncs, ncs))
            forces = jacs@forces.reshape((len(coords), ncs, 1))
            forces = forces.reshape(coords.shape)
            # raise Exception(forces.shape)
        else:
            forces = self.force_function(coords)
        return forces

    def step(self):
        forces = self._prev_forces
        if forces is None:
            forces = self.get_forces(self.coords)


        v = self.velocities
        coords = self.coords + v * self.dt + forces / (2 * self._mass) * self.dt**2
        if self._atomic_structs:
            com = np.tensordot(self.masses, coords, axes=[0, -2]) / np.sum(self._mass)
            coords = coords - com[:, np.newaxis, :] # don't let COM move
        forces_new = self.get_forces(coords)
        vels = v + self.dt * (forces + forces_new) / (2 * self._mass)

        self._prev_forces = forces_new
        self.velocities = vels
        self.coords = coords
        self.steps += 1

        return coords, vels, forces_new

    @staticmethod
    def _ke(_mass, v):
        # _mass and v are 2D b.c. atomic structs
        return 1 / 2 * np.sum(np.sum(_mass * v ** 2, axis=-1), axis=-1)
    def propagate(self, num_steps=1):

        for _ in range(num_steps):
            c, v, f = self.step()
            if self.steps % self.sampling_rate == 0:
                if self.kinetic_energies is not None:
                    self.kinetic_energies.append(self._ke(self._mass, v))
                self.trajectory.append(c)
                if self.velocity_deque is not None:
                    self.velocity_deque.append(v)

        return self.trajectory

    def build_interpolation(self, energy_function, interpolation_order=2,
                            equilibration_steps=None,
                            interpolator_class=None, eckart_embed=True, reference=None, **interpolator_options):

        traj = np.array(self.trajectory)
        if equilibration_steps is not None:
            traj = traj[equilibration_steps:]
        traj = traj.reshape((-1,) + traj.shape[-2:])

        vals = [energy_function(traj)]

        # if clustering_criterion > ...:
        #     ...

        # TODO: add clustering radius + energy cutoff within disks

        if interpolation_order > 0:
            vals.append(-self.force_function(traj))
        if interpolation_order > 1:
            cshape = (-1,) + traj.shape[-2:]
            npts = np.prod(traj.shape[-2:], dtype=int)
            grad = lambda x:-self.force_function(x.reshape(cshape)).reshape(x.shape)
            hess_fun = FiniteDifferenceDerivative(grad, function_shape=(npts, npts))
            hess = np.moveaxis(
                hess_fun.derivatives(traj.reshape(-1, npts)).derivative_tensor([1])[0],
                1, 0
            )
            vals.append(hess)
        if interpolation_order > 2:
            raise ValueError("can only do order 2 interps for now because I am laaazy")

        if eckart_embed:
            if isinstance(reference, Molecule):
                ref = reference
            elif reference is None:
                ref_pos = np.argmin(vals[0])
                ref = Molecule(
                    ["H"] * len(self.masses),
                    traj[ref_pos],
                    masses=self.masses,
                )
            else:
                ref = Molecule(
                    ["H"] * len(self.masses),
                    reference,
                    masses=self.masses,
                )
            ref = ref.get_embedded_molecule(load_properties=False)

            # all_crds = []
            # all_ders = [[] for _ in range(interpolation_order)]

            rots, _, (pax_traj, _, pax_rots) = ref.get_embedding_data(traj)
            traj = pax_traj @ np.swapaxes(rots, -2, -1)

            if interpolation_order > 0:
                npts = np.prod(traj.shape[-2:], dtype=int)
                cshape = traj.shape[-2:]

                # new_derivs = [vals[1] @ pax_rots @ np.swapaxes(rots, -2, -1)]
                # raise Exception(new_derivs[0])
                new_derivs = PropertyManager._transform_derivatives(
                    [v.reshape((-1,) + (npts,)*(i+1)) for i,v in enumerate(vals[1:])],
                    rots @ pax_rots
                )
                vals = [vals[0]] + [v.reshape((-1,) + cshape*(i+1)) for i,v in enumerate(new_derivs)]


            # for crd_ders in zip(traj, *vals[1:]): #TODO: speed this up...
            #     c = crd_ders[0]
            #     mol = Molecule(
            #         ["H"] * len(self.masses),
            #         c,
            #         masses=self.masses,
            #         potential_derivatives=crd_ders[1:]
            #     ).get_embedded_molecule(ref, load_properties=False)
            #     all_crds.append(mol.coords)
            #     pes = mol.potential_derivatives
            #     for i, p in enumerate(pes):
            #         all_ders[i].append(p)

            # traj = np.array(all_crds)

        if interpolator_class is None:
            interpolator_class = RBFDInterpolator

        return interpolator_class(traj, *vals, **interpolator_options)

    def extract_trajectory(self, flatten=True, embed=True, extract_velocities=None):
        ref_struct = None
        if isinstance(embed, (np.ndarray, list, tuple)):
            ref_struct = np.asanyarray(embed)
            embed = True
        base_coords = np.array(self.trajectory)
        if flatten:
            base_coords = base_coords.reshape((-1,)+base_coords.shape[-2:])
        if extract_velocities is None:
            extract_velocities = self.velocity_deque is not None
        if extract_velocities:
            velocities = np.array(self.velocity_deque)
        else:
            velocities = None
        if embed:
            if ref_struct is None:
                if not flatten:
                    ref_crds = base_coords.reshape((-1,)+base_coords.shape[-2:])
                else:
                    ref_crds = base_coords
                if self.kinetic_energies is not None:
                    # assuming symplectic integration...
                    kes = np.array(self.kinetic_energies).flatten()
                    max_pos = np.argmax(kes)
                    ref_struct = base_coords[max_pos]
                else:
                    ref_struct = ref_crds[0]
            mol = Molecule(["H"]*len(self.masses), ref_struct, masses=self.masses)
            if extract_velocities:
                embedding = mol.get_embedding_data(base_coords)
                coords = np.tensordot(
                    nput.vec_tensordot(
                        embedding.coord_data.coords,
                        embedding.rotations,
                        shared=1,
                        axes=[-1, 2]
                    ),
                    embedding.reference_data.axes,
                    axes=[-1, 1]
                )
                velocities = np.tensordot(
                    nput.vec_tensordot(
                        np.reshape(velocities, coords.shape),  # dx/dt has same embedding
                        embedding.rotations,
                        shared=1,
                        axes=[-1, 2]
                    ),
                    embedding.reference_data.axes,
                    axes=[-1, 1]
                )
                coords = np.reshape(coords, base_coords.shape)
                velocities = np.reshape(velocities, base_coords.shape)
            else:
                coords = mol.embed_coords(base_coords)
        else:
            coords = base_coords
        if extract_velocities:
            return coords, velocities
        else:
            return coords



class PairwisePotential:
    def __init__(self, fun, deriv=None):
        self.fun = fun
        if deriv is None:
            if hasattr(deriv, 'deriv'):
                fun.deriv()
            else:
                deriv = self.numerical_deriv(fun)
        self.deriv = deriv
        self._inds = None # to cache indices when applied to same-shape data

    @classmethod
    def numerical_deriv(cls, fun, step_size=.01):
        @functools.wraps(fun)
        def deriv(dists):
            return (fun(dists + step_size) - fun(dists - step_size)) / (2 * step_size)
        return deriv

    def _get_inds_cached(self, coords):
        if self._inds is not None and len(coords) != self._inds[0]:
            self._inds = None
        if self._inds is None:
            row_inds, col_inds = np.triu_indices(len(coords), k=1)
            _, row_map = np.unique(row_inds, return_index=True) # One element less than the coords
            sorting = np.argsort(col_inds)
            _, col_map = np.unique(col_inds, return_index=True) # One element less than the coords
            self._inds = [len(coords), (row_inds, col_inds, row_map, sorting, col_map)]
        return self._inds[1]

    def eval(self, coords):
        row_inds, col_inds, row_map, col_sort, col_map = self._get_inds_cached(coords)
        pot = self.fun

        # compute unsigned force for each pair
        diffs = coords[row_inds] - coords[col_inds]
        dists = np.linalg.norm(diffs, axis=1)
        pot_vals = pot(dists)

        # print(row_map)

        # now reduce over different chunks for the positive contribs...
        pos_chunks = np.add.reduceat(
            pot_vals,
            row_map
        )
        neg_chunks = np.add.reduceat(
            pot_vals[col_sort],
            col_map
        )

        return np.concatenate(
            [
                pos_chunks[:1],
                pos_chunks[1:] + neg_chunks[:-1],
                pos_chunks[-1:]
            ],
            axis=0
        )

    def forces(self, coords):
        row_inds, col_inds, row_map, col_sort, col_map = self._get_inds_cached(coords)
        pot_deriv = self.deriv

        # compute unsigned force for each pair
        diffs = coords[row_inds] - coords[col_inds]
        dists = np.linalg.norm(diffs, axis=1)
        base_force = pot_deriv(dists)
        normals = diffs / dists[:, np.newaxis]
        force_list = -normals * base_force[:, np.newaxis]

        # now reduce over different chunks for the positive contribs...
        pos_chunks = np.add.reduceat(
            force_list,
            row_map
        )
        neg_chunks = np.add.reduceat(
            force_list[col_sort],
            col_map
        )

        return np.concatenate(
            [
                pos_chunks[:1],
                pos_chunks[1:] - neg_chunks[:-1],
                pos_chunks[-1:]
            ],
            axis=0
        )

    def __call__(self, coords):
        return self.eval(coords)




import enum, numpy as np, functools as fp

from McUtils.Scaffolding import Logger, NullLogger, Checkpointer, NullCheckpointer
from McUtils.Parallelizers import Parallelizer
from McUtils.Zachary import TensorDerivativeConverter, TensorExpansionTerms

class JacobianKeys(enum.Enum):
    CartesiansByInternals = "CartesiansByInternals"
    InternalsByCartesians = "InternalsByCartesians"
    InternalsByCartesianModes = "InternalsByModes"
    CartesianModesByInternals = "ModesByInternals"
    CartesiansByInternalModes = "CartesiansByModes"
    InternalModesByCartesians = "ModesByCartesians"
    CartesianModesByInternalModes = "CartesianModesByInternalModes"
    InternalModesByCartesianModes = "InternalModesByCartesianModes"
    InternalModesByInternals = "InternalModesByInternals"
    InternalsByInternalModes = "InternalsByInternalModes"
    CartesianModesByCartesians = "CartesianModesByCartesians"
    CartesiansByCartesianModes = "CartesiansByCartesianModes"


class DumbTensor:
    """
    A wrapper to make tensor algebra suck less
    """

    def __init__(self, tensor):
        self.t = tensor
    @property
    def shape(self):
        return self.t.shape
    @staticmethod
    def _dot(*t, axes=None):
        """
        Flexible tensordot
        """

        if len(t) == 1:
            return t[0]

        if any(isinstance(x, int) for x in t):
            return 0

        def tdot(a, b, **kw):
            if hasattr(a, "tensordot"):
                if 'axes' not in kw:
                    kw['axes'] = [-1, 0]
                td = a.tensordot(b, **kw)
            else:
                try:
                    td = np.tensordot(a, b, **kw)
                except ValueError:
                    if 'axes' not in kw:
                        axes = [-1, 0]
                    else:
                        axes = kw['axes']
                    raise ValueError("Shape-mismatch for sum: {} x {} along axes {}".format(a.shape, b.shape, axes))
            return td

        def td(a, b):
            if isinstance(a, int) or isinstance(b[0], int):
                res = 0
            else:
                res = tdot(a, b[0], axes=b[1])
            return res

        if axes is None:
            axes = [1] * (len(t) - 1)


        return fp.reduce(td, zip(t[1:], axes), t[0])

    def dot(self, b, *args, **kwargs):
        if isinstance(b, DumbTensor):
            b = b.t
        return type(self)(self._dot(self.t, b, *args, **kwargs))

    @staticmethod
    def _shift(a, *s):
        if isinstance(a, int):
            return a

        def shift_inds(n, i, j):
            if i < j:
                x = list(range(i)) + list(range(i + 1, j + 1)) + [i] + list(range(j + 1, n))
            else:
                x = list(range(j)) + [i] + list(range(j, i)) + list(range(i + 1, n))
            return x

        shiftIJ = lambda a, ij: np.transpose(a, shift_inds(a.ndim, *ij))
        return fp.reduce(shiftIJ, s, a)
    def shift(self, *args, **kwargs):
        return type(self)(self._shift(self.t, *args, **kwargs))
    def transpose(self, *perm):
        return type(self)(self.t.transpose(perm))

    @staticmethod
    def _contract_dim(R, targ_dim):
        # we figure out how much we're off by
        # and go from there, assuming that pairs of
        # dimensions to be contracted show up at the end
        for i in range(R.ndim - targ_dim):
            l_pos = R.ndim - (i + 2)
            gloobers = R.shape[:l_pos]
            if i > 0:
                r_pos = -i
                groobers = R.shape[r_pos:]
            else:
                groobers = ()
            R = R.reshape(gloobers + (-1,) + groobers)
        return R
    def contract_dim(self, targ_dim):
        return type(self)(self._contract_dim(self.t, targ_dim))

    def __add__(self, other):
        if isinstance(other, DumbTensor):
            other = other.t
        return type(self)(self.t+other)
    def __radd__(self, other):
        if isinstance(other, DumbTensor):
            other = other.t
        return type(self)(self.t+other)
    def __matmul__(self, other):
        return self.dot(other)
    def __getitem__(self, item):
        """
        :type item: slice
        """
        a = item.start
        b = item.stop
        return self.shift([a, b])

class MolecularFunctionExpander:
    """
    Handles all necessary details to be able to
    evaluate functions in molecular internal coordinates
    and then transform back to Cartesians, and to do the
    same with respect to normal dimensionless normal mode coordinates
    """

    def __init__(self,
                 cartesian_coordinates,
                 internal_coordinates,
                 transform=None,
                 masses=None, freqs=None,

                 parallelizer=None, logger=None,
                 checkpointer=None,

                 numerical_jacobians=True,
                 eckart_embed_derivatives=True,
                 eckart_embed_planar_ref_tolerance=None,
                 strip_dummies=False,
                 strip_embedding=True,
                 internal_fd_mesh_spacing=1.0e-3,
                 internal_fd_stencil=None,
                 cartesian_fd_mesh_spacing=1.0e-2,
                 cartesian_fd_stencil=None,
                 cartesian_analytic_deriv_order=0,
                 internal_by_cartesian_order=3,
                 cartesian_by_internal_order=4,
                 jacobian_warning_threshold=1e4
                 ):
        self._cached_transforms = {}

        self.cartesian_coordinates = cartesian_coordinates
        self.internal_coordinates = internal_coordinates
        self.modes = transform
        self.masses = np.ones(cartesian_coordinates.shape[-2]) if masses is None else masses
        self.freqs = np.ones(cartesian_coordinates.shape[-2]) if freqs is None else freqs
        if self.modes is not None:
            self.modes, self.inverse = self.undimensionalize(self.masses, self.freqs, transform)
        else:
            self.inverse = None

        self._internal_jacobians = []
        self._cartesian_jacobians = []
        self.parallelizer = parallelizer
        self.logger = logger

        self.internal_fd_mesh_spacing = internal_fd_mesh_spacing
        self.internal_fd_stencil = internal_fd_stencil
        self.cartesian_fd_mesh_spacing = cartesian_fd_mesh_spacing
        self.cartesian_fd_stencil = cartesian_fd_stencil
        self.cartesian_analytic_deriv_order = cartesian_analytic_deriv_order

        self.internal_by_cartesian_order = internal_by_cartesian_order
        self.cartesian_by_internal_order = cartesian_by_internal_order
        self.jacobian_warning_threshold = jacobian_warning_threshold

        self.strip_dummies = strip_dummies
        self.strip_embedding = strip_embedding

        self.reembed = eckart_embed_derivatives
        self.reembed_tol = eckart_embed_planar_ref_tolerance
        self.all_numerical = numerical_jacobians

        if logger is None:
            logger = NullLogger()
        self.logger = logger
        if parallelizer is None:
            parallelizer = Parallelizer.lookup(None)
        self.parallelizer = parallelizer
        if checkpointer is None:
            checkpointer = NullCheckpointer()
        self.checkpointer = checkpointer

    def _tripmass(self, masses):
        if self.strip_dummies:
            masses = masses[masses > 0]
        return np.broadcast_to(masses[np.newaxis, :], (3, len(masses))).T.flatten()
    def undimensionalize(self, masses, freqs, modes):
        """
        Removes units from normal modes

        :param masses:
        :type masses:
        :param modes:
        :type modes:
        :return:
        :rtype:
        """
        L = modes
        freq_conv = np.sqrt(np.broadcast_to(freqs[:, np.newaxis], L.shape))
        mass_conv = np.sqrt(np.broadcast_to(self._tripmass(masses)[np.newaxis, :], L.shape))
        conv = freq_conv * mass_conv
        L = L * conv
        Linv = (L / freq_conv**2)
        return L.T, Linv

    def get_int_jacobs(self, jacs):
        """
        Gets the specified Internal->Cartesian Jacobians

        :param jacs:
        :type jacs:
        :return:
        :rtype:
        """
        intcds = self.internal_coordinates
        ccoords = self.cartesian_coordinates
        carts = ccoords.system
        internals = intcds.system
        exist_jacs = self._internal_jacobians
        max_jac = max(jacs)
        need_jacs = [x+1 for x in range(0, max_jac) if x >= len(exist_jacs) or exist_jacs[x] is None]
        if len(need_jacs) > 0:
            stencil = (max(need_jacs) + 2 + (1+max(need_jacs))%2) if self.internal_fd_stencil is None else self.internal_fd_stencil
            # odd behaves better
            with Parallelizer.lookup(self.parallelizer) as par:
                new_jacs = [
                    x.squeeze() if isinstance(x, np.ndarray) else x
                    for x in intcds.jacobian(carts, need_jacs,
                                             # odd behaves better
                                             mesh_spacing=self.internal_fd_mesh_spacing,
                                             stencil=stencil,
                                             all_numerical=self.all_numerical,
                                             converter_options=dict(
                                                 reembed=self.reembed,
                                                 planar_ref_tolerance=self.reembed_tol,
                                                 strip_dummies=self.strip_dummies
                                             ),
                                             parallelizer=par
                                             )]
                # np.set_printoptions
                # with np.printoptions(linewidth=1e8, threshold=1e8, floatmode='fixed', precision=10):
                #     raise Exception(str(np.round(new_jacs[0].reshape(9, 9)[(3, 6, 7), :], 12)))
            for j,v in zip(need_jacs, new_jacs):
                for d in range(j-len(exist_jacs)):
                    exist_jacs.append(None)
                exist_jacs[j-1] = v

        return [exist_jacs[j-1] for j in jacs]
    def get_cart_jacobs(self, jacs):
        """
        Gets the specified Cartesian->Internal Jacobians

        :param jacs:
        :type jacs:
        :return:
        :rtype:
        """
        intcds = self.internal_coordinates
        ccoords = self.cartesian_coordinates
        carts = ccoords.system
        internals = intcds.system
        exist_jacs = self._cartesian_jacobians
        max_jac = max(jacs)
        need_jacs = [x+1 for x in range(0, max_jac) if x >= len(exist_jacs) or exist_jacs[x] is None]
        if len(need_jacs) > 0:
            stencil = (max(need_jacs) + 2 + (1+max(need_jacs))%2) if self.cartesian_fd_stencil is None else self.cartesian_fd_stencil
            # odd behaves better
            with Parallelizer.lookup(self.parallelizer) as par:
                new_jacs = [
                    x.squeeze() if isinstance(x, np.ndarray) else x
                    for x in ccoords.jacobian(internals, need_jacs,
                                                          mesh_spacing=self.cartesian_fd_mesh_spacing,
                                                          stencil=stencil,
                                                          # all_numerical=True,
                                                          analytic_deriv_order=self.cartesian_analytic_deriv_order,
                                                          converter_options=dict(strip_dummies=self.strip_dummies),
                                                          parallelizer=par
                                                          )
                ]
                if need_jacs[0] > self.cartesian_analytic_deriv_order:
                    new_jacs = new_jacs[self.cartesian_analytic_deriv_order:]

                for j, v in zip(need_jacs, new_jacs):
                    for d in range(j - len(exist_jacs)):
                        exist_jacs.append(None)
                    exist_jacs[j - 1] = v

        return [exist_jacs[j-1] for j in jacs]

    def _get_embedding_coords(self):
        try:
            embedding = self.internal_coordinates.system.embedding_coords
        except AttributeError:
            try:
                embedding = self.internal_coordinates.system.converter_options['embedding_coords']
            except KeyError:
                embedding = None
        return embedding
    def get_coordinate_transforms(self,
                                  internal_by_cartesian_order=None,
                                  cartesian_by_internal_order=None
                                  ):

        if internal_by_cartesian_order is None:
            internal_by_cartesian_order = self.internal_by_cartesian_order
        if cartesian_by_internal_order is None:
            cartesian_by_internal_order = self.cartesian_by_internal_order

        current_cache = self._cached_transforms

        if (
                current_cache is None
                or len(current_cache[JacobianKeys.CartesiansByInternals]) < cartesian_by_internal_order
                or len(current_cache[JacobianKeys.InternalsByCartesians]) < internal_by_cartesian_order
        ):

            if current_cache is None:
                current_cache = {}

            embedding_coords = self._get_embedding_coords() if self.strip_embedding else None
            # fill out
            if (
                    JacobianKeys.CartesiansByInternals not in current_cache
                    or len(current_cache[JacobianKeys.CartesiansByInternals]) < cartesian_by_internal_order
            ):
                cart_by_internal_jacobs = self.get_int_jacobs(list(range(1, cartesian_by_internal_order+1)))

                # The finite difference preserves too much shape by default
                _contract_dim = DumbTensor._contract_dim
                _ = []
                for i,x in enumerate(cart_by_internal_jacobs):
                    if isinstance(x, int) or x.ndim == 2+i:
                        _.append(x)
                    elif x.ndim > 2+i:
                        _.append(_contract_dim(x, 2+i))
                    else:
                        raise ValueError("bad shape for Cartesian by internal jacobian {} ({})".format(
                            i, x.shape
                        ))
                cart_by_internal_jacobs = _

                # we'll strip off the embedding coords just in case
                if embedding_coords is not None:
                    good_coords = np.setdiff1d(np.arange(3*len(self.masses)), embedding_coords)

                for i, x in enumerate(cart_by_internal_jacobs):
                    bad_spots = np.where(np.abs(x) > self.jacobian_warning_threshold)
                    bad_bad_spots = bad_spots  # so we don't lose it
                    if len(bad_spots) > 0:  # numpy fuckery
                        bad_spots = bad_spots[0]
                    if len(bad_spots) > 0:
                        m = np.max(np.abs(x[bad_bad_spots]))
                        self.logger.log_print('WARNING: maximum d^{i}X/dR^{i} term is {m}. '
                                              'This will likely mess up G-matrix terms and is probably coming from a planar structure. '
                                              'Setting to zero, but `jacobian_warning_threshold` can be increased if this is expected '
                                              'All terms >{t} (base shape:{s}): {b}',
                                              i=i+1,
                                              m=m,
                                              s=x.shape,
                                              b=np.array(bad_bad_spots).T.tolist(),
                                              t=self.jacobian_warning_threshold
                                              )
                        x[bad_bad_spots] = 0.
                        # raise Exception(";_;")


                # Need to then mass weight
                masses = self.masses
                mass_conv = np.sqrt(self._tripmass(masses))
                # mass weight the derivs w.r.t internals
                internal_weighting = mass_conv
                _ = []
                for i, x in enumerate(cart_by_internal_jacobs):
                    internal_weighting = np.expand_dims(internal_weighting, 0)
                    if isinstance(x, int):
                        _.append(x)
                    else:
                        x = x * internal_weighting
                        if embedding_coords is not None:
                            for j in range(i+1):
                                x = np.take(x, good_coords, axis=j)
                        _.append(x)
                cart_by_internal_jacobs = _

                current_cache[JacobianKeys.CartesiansByInternals] = cart_by_internal_jacobs

            else:
                cart_by_internal_jacobs = current_cache[JacobianKeys.CartesiansByInternals]

            if (
                    JacobianKeys.InternalsByCartesians not in current_cache
                    or len(current_cache[JacobianKeys.InternalsByCartesians]) < internal_by_cartesian_order
            ):
                int_by_cartesian_jacobs = self.get_cart_jacobs(list(range(1, internal_by_cartesian_order + 1)))

                _contract_dim = DumbTensor._contract_dim
                _ = []
                for i,x in enumerate(int_by_cartesian_jacobs):
                    if isinstance(x, int) or x.ndim == 2+i:
                        _.append(x)
                    elif x.ndim > 2+i:
                        _.append(_contract_dim(x, 2+i))
                    else:
                        raise ValueError("bad shape for internal by Cartesian jacobian {} ({})".format(
                            i, x.shape
                        ))
                int_by_cartesian_jacobs = _

                # we'll strip off the embedding coords just in case
                if embedding_coords is not None:
                    good_coords = np.setdiff1d(np.arange(3*len(self.masses)), embedding_coords)

                for i,x in enumerate(int_by_cartesian_jacobs):
                    bad_spots = np.where(np.abs(x) > self.jacobian_warning_threshold)
                    bad_bad_spots = bad_spots # so we don't lose it
                    if len(bad_spots) > 0: # numpy fuckery
                        bad_spots = bad_spots[0]
                    if len(bad_spots) > 0:
                        m = np.max(np.abs(x[bad_bad_spots]))
                        self.logger.log_print('WARNING: maximum d^{i}R/dX^{i} term is {m}. '
                                              'This will likely mess up G-matrix terms and is probably coming from a planar structure. '
                                              'Setting to zero, but `jacobian_warning_threshold` can be increased if this is expected. '
                                              'All terms >{t} (base shape:{s}): {b}',
                                              i=i+1,
                                              m=m,
                                              s=x.shape,
                                              b=np.array(bad_bad_spots).T,
                                              t=self.jacobian_warning_threshold
                                              )
                        x[bad_bad_spots] = 0.

                # Need to then mass weight
                masses = self.masses
                mass_conv = np.sqrt(self._tripmass(masses))
                # mass weight the derivs w.r.t cartesians
                cartesian_weighting = mass_conv
                mc = mass_conv
                _ = []
                for i, x in enumerate(int_by_cartesian_jacobs):
                    cartesian_weighting = np.expand_dims(cartesian_weighting, -1)#[..., np.newaxis]
                    if isinstance(x, int):
                        _.append(x)
                    else:
                        x = x / cartesian_weighting
                        if embedding_coords is not None:
                            x = np.take(x, good_coords, axis=-1)
                        _.append(x)
                    mc = np.expand_dims(mc, 0)
                    cartesian_weighting = cartesian_weighting * mc
                int_by_cartesian_jacobs = _

                current_cache[JacobianKeys.InternalsByCartesians] = int_by_cartesian_jacobs
            else:
                int_by_cartesian_jacobs = current_cache[JacobianKeys.InternalsByCartesians]

            QY = self.modes  # derivatives of Q with respect to the Cartesians
            YQ = self.inverse # derivatives of Cartesians with respect to Q

            if QY is not None:

                if (
                        JacobianKeys.InternalsByCartesianModes not in current_cache
                        or len(current_cache[JacobianKeys.InternalsByCartesianModes]) < internal_by_cartesian_order
                ):
                    RQ_derivs = TensorDerivativeConverter(
                        [YQ] + [0]*(len(int_by_cartesian_jacobs) - 1),
                        int_by_cartesian_jacobs
                    ).convert(order=len(int_by_cartesian_jacobs))#, check_arrays=True)
                    current_cache[JacobianKeys.InternalsByCartesianModes] = RQ_derivs
                else:
                    RQ_derivs = current_cache[JacobianKeys.InternalsByCartesianModes]

                if (
                        JacobianKeys.CartesianModesByInternals not in current_cache
                        or len(current_cache[JacobianKeys.CartesianModesByInternals]) < cartesian_by_internal_order
                ):
                    QR_derivs = TensorDerivativeConverter(
                        cart_by_internal_jacobs,
                        [QY] + [0]*(len(cart_by_internal_jacobs) - 1)
                    ).convert(order=len(cart_by_internal_jacobs))
                    current_cache[JacobianKeys.CartesianModesByInternals] = QR_derivs
                else:
                    QR_derivs = current_cache[JacobianKeys.CartesianModesByInternals]

                if (
                        JacobianKeys.CartesiansByInternalModes not in current_cache
                        or len(current_cache[JacobianKeys.CartesiansByInternalModes]) < len(cart_by_internal_jacobs)
                ):
                    x_derivs = cart_by_internal_jacobs#(YR, YRR, YRRR, YRRRR)
                    Q_derivs = RQ_derivs[:1] + [0]*(len(cart_by_internal_jacobs) - 1)
                    YQ_derivs = TensorDerivativeConverter(Q_derivs, x_derivs,
                                                          jacobians_name='Q',
                                                          values_name='X'
                                                          ).convert(order=len(cart_by_internal_jacobs))#, check_arrays=True)
                    # self._get_tensor_derivs(
                    #     YQ_derivs, (QY, 0, 0, 0),
                    #     mixed_XQ=False
                    # )

                    current_cache[JacobianKeys.CartesiansByInternalModes] = YQ_derivs

                if (
                        JacobianKeys.CartesianModesByInternalModes not in current_cache
                        or len(current_cache[JacobianKeys.CartesianModesByInternalModes]) < len(cart_by_internal_jacobs)
                ):
                    # modes = self.molecule.translation_rotation_modes[1]
                    # # raise Exception(
                    # #     modes @ modes.T,
                    # #     modes.T @ modes,
                    # #     self.molecule.translation_rotation_modes[1].shape,
                    # #     self.modes.inverse.shape
                    # # )
                    # YQ2 = np.concatenate([
                    #     modes,
                    #     self.modes.inverse  # derivatives of Cartesians with respect to Q
                    # ], axis=0)

                    YQ_derivs = current_cache[JacobianKeys.CartesiansByInternalModes]
                    qQ_derivs = TensorDerivativeConverter(YQ_derivs,
                                                          [QY] + [0] * (len(cart_by_internal_jacobs) - 1),
                                                          jacobians_name='YQ',
                                                          values_name='qY'
                                                          ).convert(order=len(cart_by_internal_jacobs))#, check_arrays=True)
                    # raise Exception(qQ_derivs[0][6:, 6:])
                    current_cache[JacobianKeys.CartesianModesByInternalModes] = qQ_derivs

                if (
                        JacobianKeys.InternalModesByCartesians not in current_cache
                        or len(current_cache[JacobianKeys.InternalModesByCartesians]) < len(int_by_cartesian_jacobs)
                ):
                    QR = QR_derivs[0]
                    QY_derivs = TensorDerivativeConverter(int_by_cartesian_jacobs,
                                                          [QR] + [0]*(len(int_by_cartesian_jacobs) - 1)
                                                          ).convert(order=len(int_by_cartesian_jacobs))#, check_arrays=True)
                    current_cache[JacobianKeys.InternalModesByCartesians] = QY_derivs

                if (
                        JacobianKeys.InternalModesByCartesianModes not in current_cache
                        or len(current_cache[JacobianKeys.InternalModesByCartesianModes]) < len(int_by_cartesian_jacobs)
                ):
                    RQ_derivs = current_cache[JacobianKeys.InternalsByCartesianModes]
                    QR = QR_derivs[0]
                    Qq_derivs = TensorDerivativeConverter(RQ_derivs,
                                                          [QR] + [0] * (len(RQ_derivs) - 1),
                                                          jacobians_name='Rq',
                                                          values_name='qR'
                                                          ).convert(order=len(RQ_derivs))
                    current_cache[JacobianKeys.InternalModesByCartesianModes] = Qq_derivs

                if (
                        JacobianKeys.CartesianModesByCartesians not in current_cache
                        or len(current_cache[JacobianKeys.CartesianModesByCartesians]) < len(cart_by_internal_jacobs)
                ):
                    current_cache[JacobianKeys.CartesianModesByCartesians] = [self.modes.matrix] + [0]*(len(cart_by_internal_jacobs)-1)
                if (
                        JacobianKeys.CartesiansByCartesianModes not in current_cache
                        or len(current_cache[JacobianKeys.CartesiansByCartesianModes]) < len(cart_by_internal_jacobs)
                ):
                    current_cache[JacobianKeys.CartesiansByCartesianModes] = [self.modes.inverse] + [0] * (len(cart_by_internal_jacobs) - 1)

                if (
                        JacobianKeys.InternalModesByInternals not in current_cache
                        or len(current_cache[JacobianKeys.InternalModesByInternals]) < len(int_by_cartesian_jacobs)
                ):
                    YR = current_cache[JacobianKeys.CartesiansByInternals][0]
                    QY = current_cache[JacobianKeys.InternalModesByCartesians][0]
                    current_cache[JacobianKeys.InternalModesByInternals] = [YR@QY] + [0]*(len(int_by_cartesian_jacobs)-1)

                if (
                        JacobianKeys.InternalsByInternalModes not in current_cache
                        or len(current_cache[JacobianKeys.InternalsByInternalModes]) < len(int_by_cartesian_jacobs)
                ):
                    RY = current_cache[JacobianKeys.InternalsByCartesians][0]
                    YQ = current_cache[JacobianKeys.CartesiansByInternalModes][0]
                    current_cache[JacobianKeys.InternalsByInternalModes] = [YQ@RY] + [0]*(len(int_by_cartesian_jacobs)-1)

        return current_cache#self._cached_transforms[self.molecule]


    @property
    def cartesian_L_matrix(self):
        return self.get_cartesians_by_cartesian_modes(1)[0]
    def get_cartesians_by_cartesian_modes(self, order=None):
        base = self.get_coordinate_transforms(
            cartesian_by_internal_order=order,
            internal_by_cartesian_order=None if order is None else min(order, self.internal_by_cartesian_order)
        )[JacobianKeys.CartesiansByCartesianModes]
        if order is not None:
            if len(base) < order:
                raise ValueError("insufficient {} (have {} but expected {})".format(
                    'CartesiansByInternalModes',
                    len(base),
                    order
                ))
            base = base[:order]
        return base
    @property
    def cartesian_L_inverse(self):
        return self.get_cartesian_modes_by_cartesians(1)[0]
    def get_cartesian_modes_by_cartesians(self, order=None):
        base = self.get_coordinate_transforms(
            cartesian_by_internal_order=order,
            internal_by_cartesian_order=None if order is None else min(order, self.internal_by_cartesian_order)
        )[JacobianKeys.CartesianModesByCartesians]
        if order is not None:
            if len(base) < order:
                raise ValueError("insufficient {} (have {} but expected {})".format(
                    'CartesiansByInternalModes',
                    len(base),
                    order
                ))
            base = base[:order]
        return base

    @property
    def internal_L_matrix(self):
        return self.get_internal_modes_by_internals(1)[0]
    def get_internal_modes_by_internals(self, order=None, strip_embedding=True):
        # print(dict(
        #     cartesian_by_internal_order=order,
        #     internal_by_cartesian_order=min(order, self.internal_by_cartesian_order)
        # ))
        base = self.get_coordinate_transforms(
            cartesian_by_internal_order=order,
            internal_by_cartesian_order=None if order is None else min(order, self.internal_by_cartesian_order)
        )[JacobianKeys.InternalModesByInternals]
        if order is not None:
            if len(base) < order:
                raise ValueError("insufficient {} (have {} but expected {})".format(
                    'InternalModesByInternal',
                    len(base),
                    order
                ))
            base = base[:order]

        embedding_coords = self._get_embedding_coords() if (self.strip_embedding or strip_embedding) else None
        if embedding_coords is not None and (strip_embedding and not self.strip_embedding):
            good_coords = np.setdiff1d(np.arange(3 * self.num_atoms), embedding_coords)
            base = [t[good_coords,] if not isinstance(t, int) else t for t in base]
        return base
    @property
    def internal_L_inverse(self):
        return self.get_internals_by_internal_modes(1)[0]
    def get_internals_by_internal_modes(self, order=None, strip_embedding=True):
        base = self.get_coordinate_transforms(
            cartesian_by_internal_order=order,
            internal_by_cartesian_order=None if order is None else min(order, self.internal_by_cartesian_order)
        )[JacobianKeys.InternalsByInternalModes]
        if order is not None:
            if len(base) < order:
                raise ValueError("insufficient {} (have {} but expected {})".format(
                    'CartesiansByInternalModes',
                    len(base),
                    order
                ))
            base = base[:order]
        embedding_coords = self._get_embedding_coords() if (self.strip_embedding or strip_embedding) else None
        if embedding_coords is not None and (strip_embedding and not self.strip_embedding):
            good_coords = np.setdiff1d(np.arange(3 * self.num_atoms), embedding_coords)
            base = [t[..., good_coords] if not isinstance(t, int) else t for t in base]
        return base
    @property
    def cartesians_by_modes(self):
        return self.get_cartesians_by_modes()
    def get_cartesians_by_modes(self, order=None):
        # print(dict(
        #     cartesian_by_internal_order=order,
        #     internal_by_cartesian_order=min(order, self.internal_by_cartesian_order)
        # ))
        base = self.get_coordinate_transforms(
            cartesian_by_internal_order=order,
            internal_by_cartesian_order=None if order is None else min(order, self.internal_by_cartesian_order)
        )[JacobianKeys.CartesiansByInternalModes]
        if order is not None:
            if len(base) < order:
                raise ValueError("insufficient {} (have {} but expected {})".format(
                    'CartesiansByInternalModes',
                    len(base),
                    order
                ))
            base = base[:order]
        return base
    @property
    def modes_by_cartesians(self):
        return self.get_coordinate_transforms()[JacobianKeys.InternalModesByCartesians]
    def get_modes_by_cartesians(self, order=None, strip_embedding=True):
        base = self.get_coordinate_transforms(
            cartesian_by_internal_order=None if order is None else min(order, self.cartesian_by_internal_order),
            internal_by_cartesian_order=order
        )[JacobianKeys.InternalModesByCartesians]
        if order is not None:
            if len(base) < order:
                raise ValueError("insufficient {} (have {} but expected {})".format(
                    'InternalModesByCartesians',
                    len(base),
                    order
                ))
            base = base[:order]
        return base
    @property
    def cartesians_by_internals(self):
        return self.get_coordinate_transforms()[JacobianKeys.CartesiansByInternals]
    def get_cartesians_by_internals(self, order=None, strip_embedding=False):
        base = self.get_coordinate_transforms(
            cartesian_by_internal_order=order,
            internal_by_cartesian_order=None if order is None else min(order, self.internal_by_cartesian_order)
        )[JacobianKeys.CartesiansByInternals]
        if order is not None:
            if len(base) < order:
                raise ValueError("insufficient {} (have {} but expected {})".format(
                    'CartesiansByInternals',
                    len(base),
                    order
                ))
            base = base[:order]

        embedding_coords = self._get_embedding_coords() if (self.strip_embedding or strip_embedding) else None
        if embedding_coords is not None and (strip_embedding and not self.strip_embedding):
            good_coords = np.setdiff1d(np.arange(3 * len(self.masses)), embedding_coords)
            base = [t[np.ix_(*((good_coords,)*(t.ndim-1)))] for t in base]
        return base
    @property
    def internals_by_cartesians(self):
        return self.get_coordinate_transforms()[JacobianKeys.InternalsByCartesians]
    def get_internals_by_cartesians(self, order=None, strip_embedding=False):
        base = self.get_coordinate_transforms(
            cartesian_by_internal_order=None if order is None else min(order, self.cartesian_by_internal_order),
            internal_by_cartesian_order=order
        )[JacobianKeys.InternalsByCartesians]
        if order is not None:
            if len(base) < order:
                raise ValueError("insufficient {} (have {} but expected {})".format(
                    'InternalsByCartesians',
                    len(base),
                    order
                ))
            base = base[:order]
        embedding_coords = self._get_embedding_coords() if (self.strip_embedding or strip_embedding) else None
        if embedding_coords is not None and (strip_embedding and not self.strip_embedding):
            good_coords = np.setdiff1d(np.arange(3 * len(self.masses)), embedding_coords)
            base = [t[..., good_coords] for t in base]
        return base

    @property
    def cartesian_modes_by_internal_modes(self):
        return self.get_coordinate_transforms()[JacobianKeys.CartesianModesByInternalModes]
    def get_cartesian_modes_by_internal_modes(self, order=None):
        base = self.get_coordinate_transforms(
            cartesian_by_internal_order=order,
            internal_by_cartesian_order=None if order is None else min(order, self.internal_by_cartesian_order)
        )[JacobianKeys.CartesianModesByInternalModes]
        if order is not None:
            if len(base) < order:
                raise ValueError("insufficient {} (have {} but expected {})".format(
                    'CartesianModesByInternalModes',
                    len(base),
                    order
                ))
            base = base[:order]
        return base

    @property
    def internal_modes_by_cartesian_modes(self):
        return self.get_coordinate_transforms()[JacobianKeys.InternalModesByCartesianModes]
    def get_internal_modes_by_cartesian_modes(self, order=None):
        base = self.get_coordinate_transforms(
            cartesian_by_internal_order=None if order is None else min(order, self.cartesian_by_internal_order),
            internal_by_cartesian_order=order
        )[JacobianKeys.InternalModesByCartesianModes]
        if order is not None:
            if len(base) < order:
                raise ValueError("insufficient {} (have {} but expected {})".format(
                    'InternalModesByCartesianModes',
                    len(base),
                    order
                ))
            base = base[:order]
        return base

    def evaluate(self, function, deriv_order=2):
        ...




import numpy as np, os
from McUtils.Data import AtomData, UnitsData
import McUtils.Plots as plt

from .DGB import DGB
from .Coordinates import DGBCoords, DGBCartesians
from .Wavefunctions import DGBWavefunctions, DGBWavefunction

__all__ = [
    "DGBRunner"
]

__reload_hook__ = ['.DGB', '.Coordinates', '.Wavefunctions']

class DGBRunner:

    @classmethod
    def run_simple(cls,
                coords,
                pot,
                dipole,
                *,
                logger=True,
                plot_wavefunctions=True,
                plot_spectrum=True,
                **opts
        ):

            dgb = DGB.construct(
                np.round(coords, 8),  # this ends up really mattering to keep optimize_centers stable
                pot,
                logger=logger,
                **opts
            )

            logger = dgb.logger
            with logger.block(tag="Running DGB"):
                logger.log_print("num coords: {c}", c=len(dgb.gaussians.coords.centers))
                with logger.block(tag="S"):
                    logger.log_print(logger.prep_array(dgb.S[:5, :5]))
                with logger.block(tag="T"):
                    logger.log_print(logger.prep_array(dgb.T[:5, :5]))
                with logger.block(tag="V"):
                    logger.log_print(logger.prep_array(dgb.V[:5, :5]))

                wfns_cart = dgb.get_wavefunctions()
                with logger.block(tag="Energies"):
                    logger.log_print(
                        logger.prep_array(wfns_cart.energies[:5] * UnitsData.convert("Hartrees", "Wavenumbers"))
                    )
                with logger.block(tag="Frequencies"):
                    logger.log_print(
                        logger.prep_array(wfns_cart.frequencies()[:5] * UnitsData.convert("Hartrees", "Wavenumbers"))
                    )

                if plot_wavefunctions:
                    for i in range(4):
                        wfns_cart[i].plot().show()

                spec = wfns_cart[:4].get_spectrum(dipole)
                with logger.block(tag="Intensities"):
                    logger.log_print(
                        logger.prep_array(spec.intensities)
                    )
                if plot_spectrum:
                    spec.plot().show()

    @classmethod
    def plot_dgb_potential(cls,
                           dgb, mol, potential,
                           coordinate_sel=None,
                           domain=None, domain_padding=1,
                           potential_cutoff=17000,
                           potential_min=0,
                           plot_cartesians=None,
                           plot_atoms=True,
                           cmap=None,
                           modes_nearest=False,
                           plot_points=100,
                           levels=24,
                           **plot_styles
                           ):
        def cutoff_pot(points, cutoff=potential_cutoff / UnitsData.hartrees_to_wavenumbers,
                       cutmin=potential_min / UnitsData.hartrees_to_wavenumbers
                       ):
            values = potential(points)
            values[np.logical_or(values < cutmin, values > cutoff)] = cutoff
            return values

        if isinstance(dgb, DGBCoords):
            coords = dgb
        else:
            coords = dgb.gaussians.coords

        if plot_cartesians is None:
            plot_cartesians = isinstance(coords, DGBCartesians)
            if coordinate_sel is None:
                coordinate_sel = [0, 1]
        if plot_cartesians:
            figure = mol.plot_molecule_function(
                cutoff_pot,
                axes=coordinate_sel,
                domain=domain,
                modes_nearest=modes_nearest,
                domain_padding=domain_padding,
                cmap=cmap,
                plot_points=plot_points,
                levels=levels,
                plot_atoms=plot_atoms,
                **plot_styles
            )
        else:
            if domain is None:
                from McUtils.Zachary import Mesh
                domain = Mesh(coords.centers).bounding_box
            points = DGBWavefunction.prep_plot_grid(
                domain,
                domain_padding=domain_padding,
            )
            if points.shape[-1] == 1:
                figure = plt.Plot(
                    *np.moveaxis(points, -1, 0),
                    potential(points),
                    **plot_styles
                )
            else:
                figure = plt.TriContourPlot(
                    *np.moveaxis(points, -1, 0),
                    cutoff_pot(points),
                    cmap=cmap,
                    levels=levels,
                    **plot_styles
                )

        return figure

    @classmethod
    def plot_gaussians(cls,
                       dgb, mol,
                       *,
                       domain=None,
                       domain_padding=1,
                       cmap='RdBu',
                       plot_dir=None,
                       plot_name='gaussian_{i}.pdf',
                       **plot_options
                       ):
        n = len(dgb.gaussians.coords.centers)
        wfns = DGBWavefunctions(
            np.zeros(n),
            np.eye(n),
            dgb
        )
        cls.plot_wavefunctions(
            wfns, dgb, mol,
            cmap=cmap,
            plot_name=plot_name,
            plot_dir=plot_dir,
            domain=domain,
            domain_padding=domain_padding,
            potential_styles=dict(
                domain=domain,
                domain_padding=domain_padding
            ),
            **plot_options
        )

    default_num_plot_wfns = 5
    @classmethod
    def plot_wavefunctions(cls,
                           wfns, dgb, mol,
                           which=True,
                           coordinate_sel=None,
                           cartesians=None,
                           plot_dir=None,
                           plot_name='wfn_{i}.pdf',
                           plot_label='{e} cm-1',
                           plot_potential=True,
                           plot_atoms=None,
                           plot_centers=True,
                           potential_styles=None,
                           **plot_options
                           ):

        figure = None
        if cartesians:
            wfns = wfns.as_cartesian_wavefunction()
            dgb = wfns.hamiltonian

        if coordinate_sel is None:
            coordinate_sel = list(range(dgb.gaussians.alphas.shape[-1]))

        figs = []
        if which is True:
            which = cls.default_num_plot_wfns
        if isinstance(which, int):
            which = list(range(which))
        if isinstance(which, list) and len(which) > 0:
            pot = dgb.pot.potential_function
            if plot_potential:
                if plot_atoms is None:
                    plot_atoms = bool(plot_centers)
                if potential_styles is None:
                    potential_styles = {}
                pot_figure = cls.plot_dgb_potential(
                    dgb, mol, pot,
                    coordinate_sel=coordinate_sel,
                    plot_atoms=plot_atoms,
                    **potential_styles
                )
            else:
                pot_figure = None

            if plot_dir is not None:
                os.makedirs(plot_dir, exist_ok=True)
            if (
                    coordinate_sel is None and wfns.gaussians.alphas.shape[-1] == 1
                    or len(coordinate_sel) == 1
            ):
                for k in ['cmap', 'levels', 'plotter', 'contour_levels']:
                    if k in plot_options: del plot_options[k]
                    if k in potential_styles: del potential_styles[k]
            for i in which:
                if i < len(wfns):
                    if pot_figure is not None:
                        figure = pot_figure.copy()
                    w = wfns[i]
                    if i == 0:
                        e = w.energy
                    else:
                        e = w.energy - wfns.energies[0]
                    e = e * UnitsData.hartrees_to_wavenumbers
                    lab = plot_label.format(i=i, e=e) if plot_label is not None else None

                    if isinstance(dgb.gaussians.coords, DGBCartesians):

                        figs.append(
                            w.plot_cartesians(
                                coordinate_sel,
                                plot_centers=plot_centers,
                                figure=figure,
                                plot_label=lab,
                                **plot_options
                            )
                        )
                    else:
                        if wfns.gaussians.alphas.shape[-1] > 1:
                            if coordinate_sel is not None:
                                w = w.project(coordinate_sel)
                            figs.append(
                                w.plot(
                                    figure=figure,
                                    plot_centers=plot_centers,
                                    plot_label=lab,
                                    **plot_options
                                )
                            )
                        else:
                            figs.append(
                                w.plot(
                                    figure=figure,
                                    plot_centers=plot_centers,
                                    plot_label=lab,
                                    **plot_options
                                )
                            )

                    if plot_dir is not None:
                        fig = figs.pop()
                        fig.savefig(os.path.join(plot_dir, plot_name.format(i=i)))
                        fig.close()

            if plot_dir is None:
                figs[0].show()

    @classmethod
    def runDGB(cls,
               dgb: DGB,
               mol,
               plot_centers=True,
               plot_wavefunctions=True,
               plot_spectrum=False,
               pot_cmap='viridis',
               wfn_cmap='RdBu',
               wfn_points=100,
               wfn_contours=12,
               plot_dir=None,
               plot_potential=True,
               pot_points=100,
               domain=None,
               domain_padding=1,
               potential_cutoff=15000,
               mode=None,
               nodeless_ground_state=None,
               min_singular_value=None,
               subspace_size=None,
               plot_similarity=False,
               similarity_cutoff=None,
               similarity_chunk_size=None,
               similar_det_cutoff=None,
               **plot_options
               ):

        print("--->", len(dgb.gaussians.coords.centers))
        print(dgb.S[:5, :5])
        print(dgb.T[:5, :5])
        print(dgb.V[:5, :5])

        try:
            wfns, spec = dgb.run(
                calculate_spectrum=plot_spectrum,
                mode=mode,
                nodeless_ground_state=nodeless_ground_state,
                subspace_size=subspace_size,
                min_singular_value=min_singular_value,
                similarity_cutoff=similarity_cutoff,
                similarity_chunk_size=similarity_chunk_size,
                similar_det_cutoff=similar_det_cutoff
            )
        except Exception as e:

            if plot_wavefunctions is not False:
                print(e)

                if isinstance(plot_wavefunctions, str) and plot_wavefunctions == 'cartesians':
                    plot_wavefunctions = {'cartesians': None}
                cartesian_plot_axes = None
                if isinstance(plot_wavefunctions, dict):
                    if 'cartesians' in plot_wavefunctions:
                        cartesian_plot_axes = plot_wavefunctions['cartesians']
                        dgb = dgb.as_cartesian_dgb()
                    else:
                        raise ValueError(plot_wavefunctions)

                pot = dgb.pot.potential_function
                figure = cls.plot_dgb_potential(
                    dgb, mol, pot,
                    domain=domain, domain_padding=domain_padding,
                    potential_cutoff=potential_cutoff,
                )

                dgb.gaussians.plot_centers(
                    figure,
                    xyz_sel=cartesian_plot_axes
                )

                figure.show()

            raise e
        else:
            if plot_similarity:
                plt.ArrayPlot(dgb.get_similarity_matrix()).show()

            use_cartesians = False
            if isinstance(plot_wavefunctions, str) and plot_wavefunctions == 'cartesians':
                plot_wavefunctions = {'cartesians': None}
            coordinate_sel = None
            if isinstance(plot_wavefunctions, dict):
                if 'cartesians' in plot_wavefunctions:
                    use_cartesians = True
                    coordinate_sel = plot_wavefunctions['cartesians']
                    plot_wavefunctions = plot_wavefunctions.get('num', True)
                elif 'modes' in plot_wavefunctions:
                    coordinate_sel = plot_wavefunctions['modes']
                    plot_wavefunctions = plot_wavefunctions.get('num', True)
                    plot_potential = False
                else:
                    raise ValueError(plot_wavefunctions)

            if plot_spectrum:
                # which = plot_wavefunctions
                # if which is True:
                #     which = cls.default_num_plot_wfns
                # if isinstance(which, int):
                #     which = list(range(which))
                # print(which, spec.intensities, spec.frequencies, spec[which])
                if plot_dir is not None:
                    spec_plot = spec.plot()
                    os.makedirs(plot_dir, exist_ok=True)
                    spec_plot.savefig(os.path.join(plot_dir, 'spec.png'))
                    spec_plot.close()

            if plot_wavefunctions:
                cls.plot_wavefunctions(
                    wfns,
                    dgb,
                    mol,
                    which=plot_wavefunctions,
                    cartesians=use_cartesians,
                    coordinate_sel=coordinate_sel,
                    plot_dir=plot_dir,
                    contour_levels=wfn_contours,
                    cmap=wfn_cmap,
                    plot_points=wfn_points,
                    plot_centers={'color': 'red'} if plot_centers else False,
                    domain=domain,
                    domain_padding=domain_padding,
                    plot_potential=plot_potential,
                    scaling=.2,
                    plotter=plt.TriContourLinesPlot,
                    potential_styles=dict(
                        domain=domain,
                        domain_padding=domain_padding,
                        cmap=pot_cmap,
                        plot_points=pot_points
                    ),
                    **plot_options
                )

            if plot_spectrum:
                return wfns, spec
            else:
                return wfns


    @classmethod
    def getMorseParameters(cls, w=None, wx=None, m1=None, m2=None, re=None):
        if w is None:
            w = 3869.47 / UnitsData.hartrees_to_wavenumbers
        freq = w
        if wx is None:
            wx = 84 / UnitsData.hartrees_to_wavenumbers
        anh = wx
        De = (freq ** 2) / (4 * anh)
        if m1 is None:
            m1 = AtomData["O", "Mass"] * UnitsData.convert("AtomicMassUnits", "AtomicUnitOfMass")
        if m2 is None:
            m2 = AtomData["H", "Mass"] * UnitsData.convert("AtomicMassUnits", "AtomicUnitOfMass")
        muv = (1 / m1 + 1 / m2)
        a = np.sqrt(2 * anh / muv)

        if re is None:
            re = 1.82534

        return (De, a, re)

    @classmethod
    def setupMorseFunction(cls, model, i, j, w=None, wx=None):

        from McUtils.Data import PotentialData

        if isinstance(model, float):
            m1 = model
            m2 = i
            re = j
        else:
            m1 = model.vals[model.m(i)]
            m2 = model.vals[model.m(j)]
            re = model.vals[model.r(i, j)]

        De, a, re = cls.getMorseParameters(w=w, wx=wx, m1=m1, m2=m2, re=re)

        def morse_basic(r,
                        re=re,
                        alpha=a,
                        De=De,
                        deriv_order=None,
                        _morse=PotentialData["MorsePotential"]
                        ):
            return _morse(r, re=re, alpha=alpha, De=De, deriv_order=deriv_order)

        return morse_basic

    @classmethod
    def plot_interpolation_error(cls, dgb, pot):
        sel = slice(None)  # slice(15,30)

        centers = dgb.gaussians.overlap_data.centers[sel]
        embpot = dgb.gaussians.coords.embed_function(pot)
        realpots, realgrad, real_hess = [d * 219475 for d in embpot(centers, deriv_order=2)]
        interpots, intergrad, inter_hess = [d * 219475 for d in dgb.pot.potential_function(centers, deriv_order=2)]

        ords = np.argsort(realpots)
        devs = interpots[ords] - realpots[ords]
        rows, cols = np.triu_indices_from(dgb.S)
        utris = dgb.S[rows, cols]
        unscaled_devs = devs
        devs = devs * utris
        max_dev_pos = np.flip(np.argsort(np.abs(devs)))[:5]

        inter_gnorm = np.sum(np.abs(intergrad[ords]), axis=-1)
        real_gnorm = np.sum(np.abs(realgrad[ords]), axis=-1)
        dev_gnorm = inter_gnorm - real_gnorm
        # dev_hess = inter_hess - real_hess
        grad_plot = plt.ScatterPlot(realpots[ords], dev_gnorm)

        inter_trace = np.sum(np.sum(np.abs(inter_hess[ords]), axis=-1), axis=-1)
        real_trace = np.sum(np.sum(np.abs(real_hess[ords]), axis=-1), axis=-1)
        dev_trace = inter_trace - real_trace
        dev_hess = inter_hess - real_hess
        hess_plot = plt.ScatterPlot(realpots[ords], dev_trace)

        print("Mean Absolute Error:", np.mean(np.abs(unscaled_devs)), "Std:", np.std(unscaled_devs))
        print("Mean Scaled Error:", np.mean(np.abs(devs)), "Std:", np.std(devs))
        print("Mean Hessian Error:", np.mean(np.abs(dev_hess.flatten())),
              "Std:", np.std(dev_hess.flatten()),
              "Max:", np.max(np.abs(dev_hess.flatten()))
              )
        print("Mean Summed Hessian Error:", np.mean(np.abs(dev_trace.flatten())),
              "Std:", np.std(dev_trace.flatten()),
              "Max:", np.max(np.abs(dev_trace.flatten())),
              )
        print("Maximum (Scaled) Error:", devs[max_dev_pos])
        print("Maximum (Scaled) Interpolation Error:")
        for l, r, c, tt, ii, ov in zip(
                dgb.gaussians.coords.centers[rows[sel][ords[max_dev_pos]]],
                dgb.gaussians.coords.centers[cols[sel][ords[max_dev_pos]]],
                dgb.gaussians.overlap_data.centers[sel][ords[max_dev_pos]],
                realpots[ords[max_dev_pos]],
                interpots[ords[max_dev_pos]],
                utris[sel][ords[max_dev_pos]]
        ):
            print(f"Centers: {c} ({ov}) <- {l} {r}")
            print(f"  Error: {ii - tt} <- {tt} {ii}")

        bad_bad = np.abs(devs) > 50

        # center_plot=plt.ScatterPlot(
        #     centers[:, 0],
        #     centers[:, 1],
        #     c=unscaled_devs
        # )
        # center_plot = plt.ScatterPlot(
        #     centers[:, 0][bad_bad],
        #     centers[:, 1][bad_bad],
        #     c='blue'
        # )
        # plt.ScatterPlot(
        #     dgb.gaussians.coords.centers[:, 0],
        #     dgb.gaussians.coords.centers[:, 1],
        #     c='red',
        #     figure=center_plot
        # )
        # woof = plt.ScatterPlot(realpots[ords], unscaled_devs / realpots[ords])
        # dev_plot = plt.ScatterPlot(realpots[ords], unscaled_devs)
        dev_plot = plt.ScatterPlot(realpots[ords], unscaled_devs)
        # dev_plot = plt.ScatterPlot(realpots[ords], devs
        #                            # plot_range=[None, [-100, 100]]
        #                            )
        dev_plot.show()


"""
Provides an implementation of the distributed Gaussian basis method of Light and I think Hamilton?
extending an implementation by Jeremy Park
"""

__all__ = []
from .DGB import *; from .DGB import __all__ as exposed
__all__ += exposed
from .Gaussians import *; from .Gaussians import __all__ as exposed
__all__ += exposed
from .Coordinates import *; from .Coordinates import __all__ as exposed
__all__ += exposed
from .Evaluators import *; from .Evaluators import __all__ as exposed
__all__ += exposed
from .Interpolation import *; from .Interpolation import __all__ as exposed
__all__ += exposed
from .Wavefunctions import *; from .Wavefunctions import __all__ as exposed
__all__ += exposed
from .Runners import *; from .Runners import __all__ as exposed
__all__ += exposed